"""
Video Processing Functions

"""
import os
import cv2
from tqdm import tqdm
import multiprocessing as mp
import pandas as pd
from numpy import vstack
from pathlib import Path
from typing import Optional

from animl import file_management


def extract_frames(files,
                   frames: int = 5,
                   fps: Optional[int] = None,
                   out_file: Optional[str] = None,
                   out_dir: str = None,
                   file_col: str = "filepath",
                   parallel: bool = True,
                   num_workers: int = min(8, max(1, os.cpu_count() - 1))):
    """
    Extract frames from video files in a given DataFrame.
    Can sample frames based on a specified number of frames or frames per second (fps).

    Args:
        files (pd.DataFrame): DataFrame containing file paths to videos and images.
        frames (int): Number of frames to sample from each video (default is 5).
        fps (Optional[int]): Frames per second to sample from each video. If specified, overrides frames.
        out_file (Optional[str]): Path to save the extracted frames manifest as a CSV file.
        out_dir (str): Directory to save extracted frame images. If None, frames are not saved as images.
        file_col (str): Column name in the DataFrame that contains the file paths (default is "filepath").
        parallel (bool): Whether to use multiprocessing for frame extraction (default is True).
        num_workers (int): Number of worker processes to use for parallel processing (default is min(8, max(1, os.cpu_count() - 1))).

    Raises:
        ValueError: If the DataFrame does not contain the specified file column or if both fps and frames are not defined.
        FileNotFoundError: If a video file does not exist.
        AssertionError: If neither fps nor frames are defined.

    Returns:
        pd.DataFrame: A DataFrame containing the file paths and corresponding frame numbers for the extracted frames.
                      The DataFrame will have columns [file_col, "frame"].
    """
    if file_management.check_file(out_file, output_type="ImageFrames"):
        return file_management.load_data(out_file)
    if not {file_col}.issubset(files.columns):
        raise ValueError(f"DataFrame must contain '{file_col}' column.")
    if (fps is not None) and (frames is not None):
        print("If both fps and frames are defined fps will be used.")
    if (fps is None) and (frames is None):
        raise AssertionError("Either fps or frames need to be defined.")

    images = files[files[file_col].apply(
        lambda x: Path(x).suffix.lower()).isin(file_management.IMAGE_EXTENSIONS)]
    images = images.assign(frame=0)

    videos = files[files[file_col].apply(
        lambda x: Path(x).suffix.lower()).isin(file_management.VIDEO_EXTENSIONS)]

    if not videos.empty:
        video_frames = []
        if parallel:
            pool = mp.Pool(num_workers)
            output = [pool.apply(count_frames, args=(video, frames, fps)) for video in tqdm(videos[file_col])]
            output = list(filter(None, output))
            video_frames = vstack(output)
            video_frames = pd.DataFrame(video_frames, columns=[file_col, "frame"])
            video_frames['frame'] = video_frames['frame'].astype(int)
            pool.close()

        else:
            for i, video in tqdm(enumerate(videos[file_col])):
                output = count_frames(video, frames=frames, fps=fps)
                if output is not None:
                    video_frames.extend(output)

            video_frames = pd.DataFrame(video_frames, columns=[file_col, "frame"])
            video_frames['frame'] = video_frames['frame'].astype(int)
        videos = videos.merge(video_frames, on=file_col)

    allframes = pd.concat([images, videos]).reset_index(drop=True)

    if (out_file is not None):
        file_management.save_data(allframes, out_file)

    if out_dir is not None:
        for _, row in videos.iterrows():
            image = get_frame_as_image(row[file_col], frame=row['frame'])
            image_path = Path(out_dir) / f"{Path(row[file_col]).stem}_{row['frame']}.jpg"
            cv2.imwrite(str(image_path), image)

    return allframes


def count_frames(filepath, frames=5, fps=None) -> int:
    """
    Count number of frames in a video

    Args:
        filepath: path to video file
        frames: number of frames to sample
        fps: frames per second to sample

    Returns:
        frames_saved: list of frames to be extracted
    """
    if not Path(filepath).is_file():
        raise FileNotFoundError(f"Video file {filepath} does not exist")

    cap = cv2.VideoCapture(filepath)
    if not cap.isOpened():  # corrupted video
        return None

    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    if frame_count == 0:
        # print(f"Video file {filepath} has 0 frames, skipping.")
        return None

    cap.release()
    cv2.destroyAllWindows()

    frames_saved = []
    frame_capture = 0

    # select by fps
    if fps is not None:
        video_fps = cap.get(cv2.CAP_PROP_FPS)
        if video_fps == 0:
            # try to calculate fps from duration
            duration = cap.get(cv2.CAP_PROP_POS_MSEC) / 1000  # Sometimes unreliable
            if duration > 0:
                video_fps = frame_count / duration
            else:
                print(f"Could not determine video FPS, defaulting to {frames} frames uniformly sampled.")
                increment = int(frame_count / frames)
                while len(frames_saved) < frames:
                    frames_saved.append([str(filepath), frame_capture])
                    frame_capture += increment
                return frames_saved

        frames = int(frame_count / video_fps * fps)
        sampled_times = [i / fps for i in range(frames)]
        frames_saved = [min(int(round(t * video_fps)), frame_count-1) for t in sampled_times]

    # select set number of frames
    else:
        increment = int(frame_count / frames)
        while len(frames_saved) < frames:
            frames_saved.append([str(filepath), frame_capture])
            frame_capture += increment

    return frames_saved


# get specific frame of video as QImage
def get_frame_as_image(video_path, frame=0):
    """
    Given a video path, return a specific frame as an RGB image

    Args:
        video_path: path to video file
        frame: frame number to extract  (default is 0)

    Returns:
        rgb_frame: extracted frame as RGB image
    """
    cap = cv2.VideoCapture(video_path)
    cap.set(cv2.CAP_PROP_POS_FRAMES, frame)
    ret, still = cap.read()  # Read the first frame
    cap.release()

    if ret:
        rgb_frame = cv2.cvtColor(still, cv2.COLOR_BGR2RGB)
    return rgb_frame



"""
Distance metrics and pairwise distance computation functions.

This module provides functions to compute distance matrices between embedding vectors
using different metrics such as Euclidean (squared) and Cosine distances. It also includes
a batched computation function to handle large datasets efficiently.

This variant removes the torch and scipy dependencies and implements everything with NumPy.

Original script from WildMe
"""
import numpy as np


def remove_diagonal(A: np.ndarray) -> np.ndarray:
    """
    Removes the diagonal elements from a square matrix.

    Args:
        A (np.ndarray): Input square matrix of shape (N, N).

    Returns:
        np.ndarray: Matrix with diagonal elements removed, shape (N, N-1).

    Raises:
        ValueError: if input is not a 2-D square array.
    """
    A = np.asarray(A)
    if A.ndim != 2:
        raise ValueError(f"Expected 2-D array, got {A.ndim}-D")
    n, m = A.shape
    if n != m:
        raise ValueError("Input must be a square matrix")

    # create mask that is True for non-diagonal elements
    mask = ~np.eye(n, dtype=bool)
    return A[mask].reshape(n, n - 1)


def euclidean_squared_distance(input1: np.ndarray, input2: np.ndarray) -> np.ndarray:
    """
    Computes pairwise Euclidean squared distances between rows of input1 and input2.

    Args:
        input1 (np.ndarray): shape (M, D)
        input2 (np.ndarray): shape (N, D)

    Returns:
        np.ndarray: distance matrix shape (M, N) where entry (i,j) is ||input1[i] - input2[j]||^2
    """
    x = np.asarray(input1, dtype=np.float64)
    y = np.asarray(input2, dtype=np.float64)

    if x.ndim != 2 or y.ndim != 2:
        raise ValueError("Both inputs must be 2-D arrays")
    if x.shape[1] != y.shape[1]:
        raise ValueError("The feature dimension (number of columns) must match")

    # Use the identity: ||x-y||^2 = ||x||^2 + ||y||^2 - 2 x.y
    x_sq = np.sum(x * x, axis=1).reshape(-1, 1)   # (M,1)
    y_sq = np.sum(y * y, axis=1).reshape(1, -1)   # (1,N)
    cross = x.dot(y.T)                            # (M,N)
    dist = x_sq + y_sq - 2.0 * cross

    # Numerical errors can make tiny negative values; clamp to zero
    np.maximum(dist, 0.0, out=dist)
    return dist


def cosine_distance(input1: np.ndarray, input2: np.ndarray, eps: float = 1e-12) -> np.ndarray:
    """
    Computes pairwise cosine distance: 1 - cosine_similarity

    Args:
        input1 (np.ndarray): shape (M, D)
        input2 (np.ndarray): shape (N, D)
        eps (float): small constant to avoid divide-by-zero for zero vectors

    Returns:
        np.ndarray: distance matrix shape (M, N) where entry (i,j) = 1 - (x_i . y_j) / (||x_i|| * ||y_j||)
    """
    x = np.asarray(input1, dtype=np.float64)
    y = np.asarray(input2, dtype=np.float64)

    if x.ndim != 2 or y.ndim != 2:
        raise ValueError("Both inputs must be 2-D arrays")
    if x.shape[1] != y.shape[1]:
        raise ValueError("The feature dimension (number of columns) must match")

    # normalize rows
    x_norm = np.linalg.norm(x, axis=1, keepdims=True)
    y_norm = np.linalg.norm(y, axis=1, keepdims=True)

    # avoid division by zero
    x_norm = np.maximum(x_norm, eps)
    y_norm = np.maximum(y_norm, eps)

    x_unit = x / x_norm
    y_unit = y / y_norm

    sim = x_unit.dot(y_unit.T)  # cosine similarity in [-1,1]
    # Clip for numerical stability
    np.clip(sim, -1.0, 1.0, out=sim)
    return 1.0 - sim


def compute_distance_matrix(input1, input2, metric: str = 'euclidean') -> np.ndarray:
    """
    A wrapper function for computing the distance matrix.

    Args:
        input1: array-like shape (M, D)
        input2: array-like shape (N, D)
        metric (str, optional): "euclidean" or "cosine".
            Default is "euclidean".

    Returns:
        np.ndarray: distance matrix shape (M, N)
    """
    x = np.asarray(input1)
    y = np.asarray(input2)

    if x.ndim != 2 or y.ndim != 2:
        raise ValueError("Both inputs must be 2-D arrays")
    if x.shape[1] != y.shape[1]:
        raise ValueError("The feature dimension (number of columns) must match")

    if metric == 'euclidean':
        return euclidean_squared_distance(x, y)
    elif metric == 'cosine':
        return cosine_distance(x, y)
    else:
        raise ValueError(
            'Unknown distance metric: {}. '
            'Please choose either "euclidean" or "cosine"'.format(metric)
        )


def compute_batched_distance_matrix(input1: np.ndarray,
                                    input2: np.ndarray,
                                    metric: str = 'cosine',
                                    batch_size: int = 10) -> np.ndarray:
    """
    Computes the distance matrix in a batched manner to save memory.

    Args:
        input1 (np.ndarray): 2-D array of query features (M, D).
        input2 (np.ndarray): 2-D array of database features (N, D).
        metric (str): The distance metric to use: 'euclidean' or 'cosine'.
        batch_size (int): The number of rows from input1 to process at a time.

    Returns:
        np.ndarray: The computed distance matrix of shape (M, N).
    """
    x = np.asarray(input1, dtype=np.float64)
    y = np.asarray(input2, dtype=np.float64)

    if x.ndim != 2 or y.ndim != 2:
        raise ValueError("Both inputs must be 2-D arrays")
    if x.shape[1] != y.shape[1]:
        raise ValueError("The feature dimension (number of columns) must match")
    if batch_size <= 0:
        raise ValueError("batch_size must be a positive integer")

    m = x.shape[0]
    num_batches = int(np.ceil(m / batch_size))
    blocks = []
    for i in range(num_batches):
        start = i * batch_size
        end = min((i + 1) * batch_size, m)
        batch = x[start:end]
        block = compute_distance_matrix(batch, y, metric=metric)
        blocks.append(block)
    return np.vstack(blocks)



"""
Tools for splitting the data for different workflows

@ Kyra Swanson 2023
"""
import pandas as pd
import numpy as np


def get_animals(manifest: pd.DataFrame):
    """
    Pulls MD animal detections for classification

    Args:
        manifest (pd.DataFrame): DataFrame containing one row for every MD detection

    Returns:
        subset of manifest containing only animal detections
    """
    return manifest[manifest['category'].astype(int) == 1].reset_index(drop=True)


def get_empty(manifest: pd.DataFrame):
    """
    Pulls MD non-animal detections

    Args:
        manifest (pd.DataFrame): DataFrame containing one row for every MD detection

    Returns:
        otherdf: subset of manifest containing empty, vehicle and human detections
        with added prediction and confidence columns
    """
    # Removes all images that MegaDetector gave no detection for
    otherdf = manifest[manifest['category'].astype(int) != 1].reset_index(drop=True)
    otherdf['prediction'] = otherdf['category'].astype(int)

    # Numbers the class of the non-animals correctly
    if not otherdf.empty:
        otherdf['prediction'] = otherdf['prediction'].replace(2, "human")
        otherdf['prediction'] = otherdf['prediction'].replace(3, "vehicle")
        otherdf['prediction'] = otherdf['prediction'].replace(0, "empty")
        otherdf['confidence'] = otherdf['conf']
        otherdf['confidence'] = otherdf['confidence'].replace(np.nan, 1)  # correct empty conf

    else:
        otherdf = pd.DataFrame(columns=manifest.columns.values)

    return otherdf



"""
Symlink Module

Provides functions for creating, removing, and updating sorted symlinks.

@ Kyra Swanson 2023
"""
import json
import os
import pandas as pd
from typing import Optional
from shutil import copy2
from random import randrange
from pathlib import Path
from tqdm import tqdm

from animl import file_management, __version__
from animl.utils.general import convert_minxywh_to_absxyxy


def export_folders(manifest: pd.DataFrame,
                   out_dir: str,
                   out_file: Optional[str] = None,
                   label_col: str = 'prediction',
                   file_col: str = "filepath",
                   unique_name: str = 'uniquename',
                   copy: bool = False) -> pd.DataFrame:
    """
    Creates symbolic links of images into species folders.

    Args:
        manifest (DataFrame): dataframe containing images and associated predictions
        out_dir (str): root directory for species folders
        out_file (Optional[str]): if provided, save the manifest to this file
        label_col (str): column containing species labels, 
                        'category' for MD categories or 'prediction' for species labels
        file_col (str): column containing source paths
        unique_name (str): column containing unique file name
        copy (bool): if true, hard copy

    Returns:
        copy of manifest with link path column
    """
    out_dir = Path(out_dir)

    if label_col not in manifest.columns:
        raise AssertionError(f"Label column {label_col} not found in manifest.")

    if label_col == 'category':
        classes = {"0": "empty", "1": "animal", "2": "human", "3": "vehicle"}
        for i in classes.values():
            path = out_dir / str(i)
            path.mkdir(exist_ok=True)
    else:
        classes = manifest[label_col].unique()
        for i in classes:
            path = out_dir / str(i)
            path.mkdir(exist_ok=True)

    # create new column
    manifest['link'] = out_dir

    for i, row in tqdm(manifest.iterrows()):
        try:
            name = row[unique_name]
        except KeyError:
            filename = Path(row[file_col]).stem
            extension = Path(row[file_col]).suffix

            # get datetime
            if "datetime" in manifest.columns:
                reformat_date = pd.to_datetime(row['datetime'], format="%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%d_%H%M%S")
            else:
                reformat_date = '{:04}'.format(randrange(1, 10 ** 5))
            # get station
            if "station" in manifest.columns:
                station = row['station']
                name = "_".join([station, reformat_date, filename]) + extension
            else:
                name = "_".join([reformat_date, filename]) + extension

            manifest.loc[i, unique_name] = name

        if label_col == 'category':
            link = out_dir / str(classes[str(row['category'])]) / str(name)
        else:
            link = out_dir / str(row[label_col]) / str(name)

        manifest.loc[i, 'link'] = str(link)

        if not link.is_file():
            if copy:  # make a hard copy
                copy2(row[file_col], link)
            else:  # make a hard
                os.link(row[file_col], link)

    if out_file:
        manifest.to_csv(out_file, index=False)

    return manifest


def remove_link(manifest: pd.DataFrame,
                link_col: str = 'link') -> pd.DataFrame:
    """
    Deletes symbolic links of images.

    Args:
        manifest (pd.DataFrame): dataframe containing images and associated predictions
        link_col (str): column name of paths to remove

    Returns:
        manifest without link column
    """
    # delete files
    for _, row in manifest.iterrows():
        Path(row[link_col]).unlink(missing_ok=True)
    # remove column
    manifest.drop(columns=[link_col])
    return manifest


def update_labels_from_folders(manifest: pd.DataFrame,
                               export_dir: str,
                               unique_name: str = 'uniquename') -> pd.DataFrame:
    """
    Update manifest after human review of symlink directories.

    Args:
        manifest (pd.DataFrame): dataframe containing images and associated predictions
        export_dir (str): root directory for species folders
        unique_name (str): column containing unique file names

    Returns:
        manifest: dataframe with updated predictions
    """
    if unique_name not in manifest.columns:
        raise AssertionError("Manifest does not have unique names, cannot match to sorted directories.")

    print("Searching directory...")
    ground_truth = file_management.build_file_manifest(export_dir, exif=False)

    if len(ground_truth) != len(manifest):
        print(f"Warning, found {len(ground_truth)} files in link dir but {len(manifest)} files in manifest.")

    # last level should be label level
    ground_truth = ground_truth.rename(columns={'filename': unique_name})
    ground_truth['label'] = ground_truth["filepath"].apply(lambda x: Path(x).parent.name)

    return pd.merge(manifest, ground_truth[[unique_name, 'label']], on=unique_name)


def export_coco(manifest: pd.DataFrame,
                class_list: pd.DataFrame,
                out_file: str,
                info: Optional[dict] = None,
                licenses: Optional[list] = None):
    """
    Export a manifest to COCO format.

    Args:
        manifest (pd.DataFrame): dataframe containing images and associated predictions
        class_list (pd.DataFrame): dataframe containing class names and their corresponding IDs
        out_file (str): path to save the COCO formatted file
        info (Optional[dict]): info section of COCO file
        licenses (Optional[list]): licenses section of COCO file

    Returns:
        coco formatted json file saved to out_file
    """
    expected_columns = ('filepath', 'filename', 'filemodifydate', 'frame',
                        'max_detection_conf', 'category', 'conf', 'bbox_x', 'bbox_y', 'bbox_w',
                        'bbox_h', 'prediction', 'confidence')

    for s in expected_columns:
        assert s in manifest.columns, f'Expected column {s} not found in results DataFrame'

    if info is None:
        info = {'description': 'COCO Export from animl',
                'version': __version__,
                'date_created': pd.Timestamp.now().strftime("%Y/%m/%d")}

    if licenses is None:
        licenses = []

    # build categories from class list
    class_dict = {row['class']: int(row['id']) for _, row in class_list.iterrows()}
    categories = []
    for _, row in class_list.iterrows():
        category = {'id': int(row['id']),
                    'name': row['class'],
                    'supercategory': 'none'}
        categories.append(category)

    # create image id based on filepath
    manifest['image_id'] = manifest.groupby('filepath').ngroup()

    images = []
    annotations = []
    for i_row, row in manifest.iterrows():

        width = int(row['width']) if not pd.isna(row['width']) else 0
        height = int(row['height']) if not pd.isna(row['height']) else 0

        image = {'id': row['image_id'],
                 'file_name': Path(row['filepath']).name,
                 'width': width,
                 'height': height}
        images.append(image)

        # convert bbox to abs coordinates
        bbox = [row['bbox_x'], row['bbox_y'], row['bbox_w'], row['bbox_h']]
        # skip annotation if bbox is NaN
        if pd.isna(bbox).any():
            continue
        bbox = convert_minxywh_to_absxyxy(bbox, width, height)
        area = bbox[2] * bbox[3]

        # get category id
        category_id = class_dict.get(row['prediction'], -1)

        annotation = {'id': i_row,
                      'image_id': row['image_id'],
                      'category_id': category_id,
                      'frame': int(row.get('frame', 0)),
                      'bbox': bbox,
                      'area': area,
                      'iscrowd': 0}
        annotations.append(annotation)

    coco_format = {'info': info,
                   'licenses': licenses,
                   'images': images,
                   'annotations': annotations,
                   'categories': categories}

    with open(out_file, 'w') as f:
        json.dump(coco_format, f)

    return coco_format


def export_camtrapR(manifest: pd.DataFrame,
                    out_dir: str,
                    out_file: Optional[str] = None,
                    label_col: str = 'prediction',
                    file_col: str = "filepath",
                    station_col: str = 'station',
                    unique_name: str = 'uniquename',
                    copy: bool = False) -> pd.DataFrame:
    """
    Export data into sorted folders organized by station

    Args:
        - manifest (pd.DataFrame): dataframe containing images and associated predictions
        - out_dir (str): directory to export sorted images
        - out_file (Optional[str]): if provided, save the manifest to this file
        - label_col (str): column containing species labels
        - file_col (str): column containing source paths
        - station_col (str): column containing station names
        - unique_name (str): column containing unique file name
        - copy (bool): if true, hard copy
    """
    expected_columns = (file_col, station_col, label_col)
    for s in expected_columns:
        assert s in manifest.columns, f'Expected column {s} not found in results DataFrame'

    manifest['link'] = out_dir

    stations = manifest.groupby(station_col)

    for station_name, station in tqdm(stations):
        for i, row in station.iterrows():
            try:
                name = row[unique_name]
            except KeyError:
                filename = Path(row[file_col]).stem
                extension = Path(row[file_col]).suffix

                # get datetime
                if "datetime" in manifest.columns:
                    reformat_date = pd.to_datetime(row['datetime'], format="%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%d_%H%M%S")
                else:
                    reformat_date = '{:04}'.format(randrange(1, 10 ** 5))
                # get station
                if "station" in manifest.columns:
                    station = row['station']
                    name = "_".join([station, reformat_date, filename]) + extension
                else:
                    name = "_".join([reformat_date, filename]) + extension

                manifest.loc[i, unique_name] = name

            link = out_dir / str(station_name) / str(row[label_col]) / str(name)

            manifest.loc[i, 'link'] = str(link)

            if not link.is_file():
                if copy:  # make a hard copy
                    copy2(row[file_col], link)
                else:  # make a hard
                    os.link(row[file_col], link)

    if out_file:
        manifest.to_csv(out_file, index=False)

    return manifest


def export_timelapse(manifest: pd.DataFrame,
                     out_dir: str,
                     only_animl: bool = True) -> Path:
    '''
    Converts the Pandas DataFrame created by running the animl classsifier to a csv file that contains columns needed for TimeLapse conversion in later step

    Credit: Sachin Gopal Wani

    Args:
        manifest - a DataFrame that contains classifications
        out_dir - location of directory where csv files will be saved
        only_animl - A bool that confirms whether we want only animal detctions or all (animal + non-animal detection from MegaDetector + classifier)

    Returns:
        animals.csv - A csv file containing all the detection and classification information for animal detections
        non-anim.csv - A csv file containing detections of all non-animals made to be similar to animals.csv in columns
        csv_loc - Location of the stored animals csv file
    '''
    # Create directory
    Path(out_dir).mkdir(exist_ok=True)

    expected_columns = ('filepath', 'filename', 'filemodifydate', 'frame',
                        'max_detection_conf', 'category', 'conf', 'bbox_x', 'bbox_y', 'bbox_w',
                        'bbox_h', 'prediction', 'confidence')

    for s in expected_columns:
        assert s in manifest.columns, f'Expected column {s} not found in manifest DataFrame'

    # Dropping unnecessary columns (Refer to columns numbers above for expected columns - 0 indexed).
    manifest = manifest.drop(['filepath', 'filemodifydate', 'max_detection_conf'], axis=1)

    # Keep relative path only
    manifest['file'] = manifest['filename']

    # Rename column names for clarity
    manifest = manifest.rename(columns={'conf': 'detection_conf', 'prediction': 'class', 'confidence': 'classification_conf'})
    csv_loc = Path(out_dir / "timelapse_manifest.csv")
    manifest.to_csv(csv_loc, index=False)

    if only_animl:
        animals = results[results['category'] == 1]
        animals.to_csv(Path(out_dir / "animals.csv"), index=False)

    # Return the location of csv for json conversion
    return csv_loc


def export_megadetector(manifest: pd.DataFrame,
                        out_file: Optional[str] = None,
                        detector: str = 'MegaDetector v5b',
                        prompt: bool = True):
    """
    Converts the .csv file [input_file] to the MD-formatted .json file [out_file].

    If [out_file] is None, '.json' will be appended to the input file.

    # Credit goes to Dan Morris https://github.com/agentmorris/MegaDetector/tree/main
    # Adding a modified script to animl-py repo

    Args:
        manifest (pd.DataFrame): dataframe containing images and associated detections
        out_file (Optional[str]): path to save the MD formatted file
        detector (str): name of the detector used
        prompt (bool): whether to prompt before overwriting existing file

    Returns:
        None, saves a json file in MD format
    """

    detection_category_id_to_name = {'0': 'empty', '1': 'animal', '2': 'person', '3': 'vehicle'}

    if out_file is None:
        out_file = 'detections.json'

    if not {'filepath', 'category', 'conf', 'bbox_x', 'bbox_y',
            'bbox_w', 'bbox_h', 'prediction', 'confidence'}.issubset(manifest.columns):
        raise ValueError("DataFrame must contain bounding boxes and confidence.")

    classification_category_name_to_id = {}
    filename_to_results = {}

    for i_row, row in manifest.iterrows():

        if str(row['category']) == '0':
            continue

        # Is this the first detection we've seen for this file?
        if row['filepath'] in filename_to_results:
            im = filename_to_results[row['filepath']]
        else:
            im = {}
            im['detections'] = []
            im['file'] = row['filepath']
            filename_to_results[im['file']] = im

        assert isinstance(row['category'], int), 'Invalid category identifier in row {}'.format(im['file'])
        detection_category_id = str(row['category'])
        assert detection_category_id in detection_category_id_to_name, \
            'Unrecognized detection category ID {}'.format(detection_category_id)

        detection = {}
        detection['category'] = detection_category_id
        detection['conf'] = row['conf']
        detection['frame'] = int(row.get('frame', 0))
        bbox = [row['bbox_x'], row['bbox_y'], row['bbox_w'], row['bbox_h']]
        detection['bbox'] = bbox
        classification_category_name = row['prediction']

        # Have we seen this classification category before?
        if classification_category_name in classification_category_name_to_id:
            classification_category_id = \
                classification_category_name_to_id[classification_category_name]
        else:
            classification_category_id = str(len(classification_category_name_to_id))
            classification_category_name_to_id[classification_category_name] = \
                classification_category_id

        classifications = [[classification_category_id, row['confidence']]]
        detection['classifications'] = classifications

        im['detections'].append(detection)

    # ...for each row

    info = {}
    info['format_version'] = '3.0'
    info['detector'] = detector
    info['classifier'] = 'Animl'

    results = {}
    results['info'] = info
    results['detection_categories'] = detection_category_id_to_name
    results['classification_categories'] = \
        {v: k for k, v in classification_category_name_to_id.items()}
    results['images'] = list(filename_to_results.values())

    # Save the results to a JSON file
    file_management.save_json(results, out_file, prompt=prompt)



'''
Main script.

Runs full animl workflow on a given directory.
User must provide MegaDetector, Classifier, and Class list files,
otherwise will pull MDv5 and the CTL Southwest v2 models by default.

Usage example
> python -m animl /home/usr/animl-py/examples/Southwest/

OR

> python -m animl /image/dir megadetector.pt classifier.pt class_file.csv

Paths to model files must be edited to local machine.

@ Kyra Swanson, 2023
'''
from animl import pipeline
import time
import argparse
from pathlib import Path


# Start timer
start_time = time.time()

# IF RUN FROM COMMAND LINE
parser = argparse.ArgumentParser(description='Folder locations for the main script')
home = Path.cwd() / 'models'
# Create and parse arguements
parser.add_argument('imagedir_config', type=str,
                    help='Path to Image Directory or Config File')
parser.add_argument('--detector', type=str, nargs='?',
                    help='Path to MD model',
                    default=Path(home / 'md_v5a.0.0.pt'))
parser.add_argument('--classifier', type=str, nargs='?',
                    help='Path to Class model',
                    default=Path(home / 'sdzwa_southwest_v3.pt'))
parser.add_argument('--classlist', type=str, nargs='?',
                    help='Path to class list',
                    default=None)
parser.add_argument('--detect_only', '-d', action='store_true',
                    help='Run detection only, skip classification')
parser.add_argument('--sort', '-s', action='store_true',
                    help='Sort images into subfolders based on classification')
parser.add_argument('--visualize', '-v', action='store_true',
                    help='Visualize detections and classifications on images')
args = parser.parse_args()

# first argument is config file
if Path(args.imagedir_config).is_file():
    pipeline.from_config(args.imagedir_config)

# first argument is a directory
else:
    if not Path(args.detector).is_file():
        raise FileNotFoundError("MegaDetector not found, must be an .onnx file.")

    if not Path(args.classifier).is_file():
        raise FileNotFoundError("Classifier not found, must be a .onnx file.")

    if not Path(args.classlist).is_file():
        raise FileNotFoundError("Class list not found, must be a .csv file.")

    # Call the main function
    pipeline.from_paths(args.imagedir_config, args.detector, args.classifier, args.classlist,
                        sort=args.sort, visualize=args.visualize, detect_only=args.detect_only)

print(f"Pipeline took {time.time() - start_time:.2f} seconds")



"""
Plot Bounding Boxes and Save Images

Functionality to draw bounding boxes and labels provided image DataFrame.

@ Kyra Swanson 2023
"""
import cv2
import argparse
import pandas as pd
import math
import numpy as np
from pathlib import Path
from typing import Union

from animl.utils import general
from animl.file_management import IMAGE_EXTENSIONS
from animl.video_processing import get_frame_as_image



MD_COLORS = {"1": (0, 255, 0), "2": (0, 0, 255),  "3": (255, 0, 0)}
MD_LABELS = {"1": "animal", "2": "human",  "3": "vehicle"}

def plot_box(rows,
             file_col: str = "filepath",
             min_conf: Union[int, float] = 0,
             label_col=None,
             show_confidence=False,
             colors = MD_COLORS,
             detector_labels = MD_LABELS,
             return_img: bool = False):
    """
    Plot a bounding box on a given (loaded) image

    Args:
        rows (pandas.DataFrame): Row from the DataFrame containing bounding box coordinates and prediction.
            Expected columns:
            - file_col
            - 'conf': Confidence score of the detection.
            - 'bbox_x': x-coordinate of the top-left corner of the bounding box.
            - 'bbox_y': y-coordinate of the top-left corner of the bounding box.
            - 'bbox_w': width of the bounding box.
            - 'bbox_h': height of the bounding box.
            - 'prediction': Prediction label to be displayed alongside the bounding box (optional).
        file_col (str): filepath column name in the DataFrame
        min_conf (int or float): Minimum confidence threshold to plot the box
        label_col (str or None): Column name containing class to print above the box. If None, no label is printed.
        show_confidence (bool): If true, show confidence score above the box.
        colors (dict): Dictionary mapping class labels to BGR color tuples for the bounding boxes.
        detector_labels (dict): Dictionary mapping detector categories to human-readable labels.
        return_img (bool): If true, return the image array with boxes overlaid, otherwise display it using cv2.imshow.

    Returns:
        None
    """
    # If a single row is passed, convert it to a DataFrame for consistency
    if isinstance(rows, pd.Series):
        rows = pd.DataFrame([rows])
        
    if not {file_col, 'conf', 'bbox_x', 'bbox_y', 'bbox_w', 'bbox_h'}.issubset(rows.columns):
        raise ValueError(f"DataFrame must contain {file_col}, 'conf', 'bbox_x', 'bbox_y', 'bbox_w', and 'bbox_h' columns.")
    
    if colors is None:
        colors = MD_COLORS
    if detector_labels is None:
        detector_labels = MD_LABELS

    # Load the image
    path = rows.iloc[0][file_col]
    if not Path(path).is_file():
        raise FileNotFoundError(f"The file {path} does not exist.")

    if Path(path).suffix.lower() in IMAGE_EXTENSIONS:
        img = cv2.imread(rows.iloc[0][file_col])
    else:
        frame = rows.iloc[0]['frame'] if 'frame' in rows.columns else 0
        img = get_frame_as_image(rows.iloc[0][file_col], frame)
    
    height, width, _ = img.shape

    font_scale = min(width, height) * 1e-3
    thickness = math.ceil(min(width, height) *  1e-3)

    for _, row in rows.iterrows():
        # Skipping the box if the confidence threshold is not met
        if (row['conf']) < min_conf:
            continue

        # If any of the box isn't defined, jump to next one
        if np.isnan(row['bbox_x']):
            continue

        bbox = [row['bbox_x'], row['bbox_y'], row['bbox_w'], row['bbox_h']]
        xyxy = general.convert_minxywh_to_absxyxy(bbox, width, height)

        color = colors[str(int(row['category']))]
        thick = int((height + width) // 900)
        cv2.rectangle(img, (xyxy[0], xyxy[1]), (xyxy[2], xyxy[3]), color, thick)

        # Printing prediction if enabled
        if label_col:
            if label_col == "category":
                label = detector_labels[str(int(row['category']))]
            else:
                label = row[label_col]

            if show_confidence:
                if 'confidence' in row and not np.isnan(row['confidence']):
                    label += f" {row['confidence']:.2f}"
                elif 'conf' in row and not np.isnan(row['conf']):
                    label += f" {row['conf']:.2f}"

            (text_width, text_height), baseline = cv2.getTextSize(label, cv2.FONT_HERSHEY_DUPLEX, font_scale, thickness)
            # move label if it is out of the image
            label_top = max(0, xyxy[1] - text_height - baseline)
            label_bottom = xyxy[1] if label_top > 0 else xyxy[1] + text_height + baseline

            label_right = min(width, xyxy[0] + text_width)
            label_left = xyxy[0] if label_right < width else xyxy[0] - text_width

            img = cv2.rectangle(img, (label_left, label_top), (label_right, label_bottom), color, -1)

            # adjust text if label is at top of image
            text_y_pos = label_bottom - 5 if label_top > 0 else label_bottom
            cv2.putText(img, label, (label_left, text_y_pos), cv2.FONT_HERSHEY_SIMPLEX, font_scale, (0,0,0), thickness)

    if return_img:
        return img
    else:
        cv2.imshow('image', img)
        cv2.waitKey(0)
        cv2.destroyAllWindows()


def plot_all_bounding_boxes(manifest: pd.DataFrame,
                            out_dir: str,
                            file_col: str = 'filepath',
                            min_conf: Union[int, float] = 0.1,
                            label_col = False,
                            show_confidence: bool = False,
                            colors = MD_COLORS,
                            detector_labels = MD_LABELS):
    """
    This function takes the parsed dataframe output from MegaDetector, makes a copy of each image,
    plots the boxes in the new image, and saves it the specified directory.

    Args:
        manifest (Pandas DataFrame): manifest of detections
        out_dir (str): Name of the output directory
        file_col (str): Column name containing file paths
        min_conf (Optional) (Int or Float): Confidence threshold to plot the box
        label_col (Optional) (str): Column name containing label to print on box
        show_confidence (Optional) (bool): If true, show confidence score on box
        colors (Optional) (dict): Dictionary mapping class labels to BGR color tuples for the bounding boxes.
        detector_labels (Optional) (dict): Dictionary mapping detector categories to human-readable labels.

    Returns:
        None
    """
    if not {file_col}.issubset(manifest.columns):
        raise ValueError(f"DataFrame must contain '{file_col}' column.")
    
    if colors is None:
        colors = MD_COLORS
    if detector_labels is None:
        detector_labels = MD_LABELS

    # If the specified output directory does not exist, make it
    Path(out_dir).mkdir(exist_ok=True)

    # iterate through unique file paths
    manifest_filepaths = manifest.groupby(file_col)
    for filepath, detections in manifest_filepaths:
        # ouput name
        file_name_no_ext = Path(filepath).stem
        file_ext = Path(filepath).suffix

        # file is an image
        if file_ext.lower() in IMAGE_EXTENSIONS:

            img = plot_box(detections, file_col=file_col, min_conf=min_conf, 
                           label_col=label_col, show_confidence=show_confidence,
                           colors=colors, detector_labels=detector_labels, return_img=True)

                # Saving the image
            new_file_path = Path(out_dir) / f"{file_name_no_ext}_box.jpg"
            cv2.imwrite(new_file_path, img)
            cv2.destroyAllWindows()
            
        # file is a video, break up by frames
        else:
            if not {'frame'}.issubset(manifest.columns):
                raise ValueError("DataFrame must contain 'frame' column for video files.")
            
            frames = detections.groupby('frame')
            for f, frame_detections in frames:

                img = plot_box(frame_detections, file_col=file_col, min_conf=min_conf,
                               label_col=label_col, show_confidence=show_confidence,
                               colors=colors, detector_labels=detector_labels, return_img=True)

                # Saving the image
                new_file_path = Path(out_dir) / f"{file_name_no_ext}_{f}_box.jpg"
                cv2.imwrite(new_file_path, img)
                cv2.destroyAllWindows()


def plot_from_file(csv_file: str, out_dir: str, file_col: str = 'filepath'):
    """
    Read a CSV manifest file and perform box plotting on the images.

    Args:
        csv_file (str): Path to the CSV file.
        out_dir (str): Saved location  of boxed images output dir.
        file_col (str): Column name containing file paths.

    Returns:
        None
    """
    # Read the CSV file
    data = pd.read_csv(csv_file)

    # Perform box plotting for each image in the CSV file
    for i, row in data.iterrows():
        img = plot_box(row, return_img=True)
        # Save the image with boxes
        file_name_no_ext = Path(row[file_col]).stem
        file_ext = Path(row[file_col]).suffix
        if file_ext.lower() not in IMAGE_EXTENSIONS:
            file_ext = '.jpg'
        new_file_path = Path(out_dir, f"{file_name_no_ext}_{i}_{file_ext}")
        cv2.imwrite(new_file_path, img)


if __name__ == '__main__':
    # Create an argument parser
    parser = argparse.ArgumentParser(description='Plot boxes images-csv')

    # Add the CSV file and output directory arguments
    parser.add_argument('csv_file', type=str, help='Path to the CSV file')
    parser.add_argument('out_dir', type=str, help='Path to the output dir')

    # Parse the command-line arguments
    args = parser.parse_args()

    # Call the main function'
    plot_from_file(args.csv_file, args.out_dir)



"""
General utils

"""
import cv2
import numpy as np
import onnxruntime as ort


def softmax(x):
    '''
    Helper function to softmax
    '''
    return np.exp(x)/np.sum(np.exp(x), axis=1, keepdims=True)


def tensor_to_onnx(tensor, channel_last=False):
    '''
    Helper function for onnx, shifts dims to BxHxWxC
    '''
    if channel_last:
        tensor = tensor.permute(0, 2, 3, 1)  # reorder BxCxHxW to BxHxWxC

    tensor = tensor.numpy()

    return tensor


def get_onnx_device(user_set=None, quiet=False):
    """
    Get gpu if available

    #TODO: test if user picks bad cuda device
    """
    providers = ort.get_available_providers()
    if 'CUDAExecutionProvider' in providers:
        # no user input
        if user_set is None:
            if not quiet:
                print('Using available CUDA device.')
            providers = ['CUDAExecutionProvider', 'CPUExecutionProvider']     
        # user selects cpu
        elif user_set == 'cpu':
            if not quiet:
                print('CUDA is available but set to cpu by user.')
                providers = ['CPUExecutionProvider']
        # user selects cuda device and is available
        elif user_set.startswith('cuda'):
            device_number = int(user_set.split(':')[-1]) if ':' in user_set else 0
            providers = [('CUDAExecutionProvider', {'device_id': device_number}), 'CPUExecutionProvider']
            if not quiet:
                print(f'Attempting to use CUDA device: {user_set}')
        # unknown user input
        else:
            if not quiet:
                print('User-specified device unknown, using available CUDA device.')
            providers = ['CUDAExecutionProvider', 'CPUExecutionProvider']

    # cuda not available
    else:
        if user_set is not None and user_set in {'cuda', 'cuda:0', 'cuda:1', 'cuda:2', 'cuda:3'}:
            if not quiet:
                print('Warning: CUDA device specified but not available, using CPU instead.')
        providers = ['CPUExecutionProvider']
    
    return providers
 
# ==============================================================================
# COORDINATE CONVERSION
# ==============================================================================
def xyxyc2xywh(x):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] where xy1=top-left, xy2=bottom-right
    y = np.copy(x)
    y[:, 0] = (x[:, 0] + x[:, 2]) / 2  # x center
    y[:, 1] = (x[:, 1] + x[:, 3]) / 2  # y center
    y[:, 2] = x[:, 2] - x[:, 0]  # width
    y[:, 3] = x[:, 3] - x[:, 1]  # height
    return y


def xywhc2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = np.copy(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


def xywhn2xyxy(x, w=640, h=640, padw=0, padh=0):
    # Convert nx4 boxes from [x, y, w, h] normalized to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = np.copy(x)
    y[:, 0] = w * (x[:, 0] - x[:, 2] / 2) + padw  # top left x
    y[:, 1] = h * (x[:, 1] - x[:, 3] / 2) + padh  # top left y
    y[:, 2] = w * (x[:, 0] + x[:, 2] / 2) + padw  # bottom right x
    y[:, 3] = h * (x[:, 1] + x[:, 3] / 2) + padh  # bottom right y
    return y


def xyxyc2xywhn(x, w=640, h=640, clip=False, eps=0.0):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] normalized where xy1=top-left, xy2=bottom-right
    if clip:
        clip_coords(x, (h - eps, w - eps))  # warning: inplace clip
    y = np.copy(x)
    y[:, 0] = ((x[:, 0] + x[:, 2]) / 2) / w  # x center
    y[:, 1] = ((x[:, 1] + x[:, 3]) / 2) / h  # y center
    y[:, 2] = (x[:, 2] - x[:, 0]) / w  # width
    y[:, 3] = (x[:, 3] - x[:, 1]) / h  # height
    return y


def xyn2xy(x, w=640, h=640, padw=0, padh=0):
    # Convert normalized segments into pixel segments, shape (n,2)
    y = np.copy(x)
    y[:, 0] = w * x[:, 0] + padw  # top left x
    y[:, 1] = h * x[:, 1] + padh  # top left y
    return y


def xywh2xyxy(bbox):
    """
    Converts bounding boxes from xywh to xyxy format.

    Args:
        bbox (list): Bounding box coordinates in the format [x_min, y_min, width, height].

    Returns:
        list: Normalized bounding box coordinates in the format [x_min, y_min, width, height].
    """
    y = np.copy(bbox)
    y[2] = y[0] + y[2]  # bottom right x
    y[3] = y[1] + y[3]  # bottom right y
    return y

# THIS ONE
def xyxy2xywh(bbox):
    """
    Converts bounding boxes from xywh to xyxy format.

    Args:
        bbox (list): Bounding box coordinates in the format [x_min, y_min, width, height].
                     x_min,y_min are the top left corner.

    Returns:
        list: Normalized bounding box coordinates in the format [x_min, y_min, width, height].
    """
    y = np.copy(bbox)
    y[2] = y[2] - y[0]  # width
    y[3] = y[3] - y[1]  # height
    return y


def convert_minxywh_to_absxyxy(bbox, width, height):
    """
    Converts bounding box from [x_min, y_min, width, height] to [x1, y1, x2, y2] format.

    Args:
        bbox (list): Bounding box in the format [x_min, y_min, width, height].
        width (int): Width of the image.
        height (int): Height of the image.

    Returns:
        list: Bounding box in the format [x1, y1, x2, y2].
    """
    x_min, y_min, w, h = bbox
    x1 = x_min
    y1 = y_min
    x2 = x_min + w
    y2 = y_min + h

    return [int(x1 * width), int(y1 * height), int(x2 * width), int(y2 * height)]



def clip_coords(boxes, shape):
    # Clip bounding xyxy bounding boxes to image shape (height, width)
    boxes[:, [0, 2]] = boxes[:, [0, 2]].clip(0, shape[1])  # x1, x2
    boxes[:, [1, 3]] = boxes[:, [1, 3]].clip(0, shape[0])  # y1, y2


def normalize_boxes(bbox, image_sizes):
    """
    Converts absolute bounding box coordinates to relative coordinates.

    Args:
        bbox (list): Absolute bounding box coordinates.
        img_size (tuple): Image size in the format (width, height).

    Returns:
        list: Normalized bounding box coordinates.
    """
    img_height, img_width  = image_sizes
    y = np.copy(bbox)   
    y[[0,2]] = np.clip(y[[0,2]] / img_width, 0, 1)
    y[[1,3]] = np.clip(y[[1,3]] / img_height, 0, 1)
    return y


# ==============================================================================
# Augmentations
# ==============================================================================

def letterbox(im: np.ndarray, new_shape = (640, 640),
              color = (114, 114, 114),
              auto: bool = True,
              scaleFill: bool = False,
              scaleup: bool = True,
              stride: int = 32):
    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return im, ratio, (dw, dh)


def scale_letterbox(bbox, resized_shape, original_shape):
    """
    Converts bounding box coordinates from a resized, letterboxed image space
    back to the original image's coordinate space. Assumes input coordinates
    are in normalized [x_corner, y_corner, width, height] format.

    Args:
        bbox (np.ndarray): A numpy array or tensor of bounding
                                             boxes, shape (n, 4), in
                                             (x_corner, y_corner, width, height) format.
                                             Coordinates are in pixels relative
                                             to the resized/padded image.
        resized_shape (tuple): The (height, width) of the resized and
                               letterboxed image.
        original_shape (tuple): The (height, width) of the original image.

    Returns:
        np.ndarray: A numpy array of bounding boxes, shape (n, 4), with
                    coordinates in normalized (x_corner, y_corner, width, height)
                    format.
    """
    # Convert input xywh (top-left corner) to xyxy
    xyxy_coords = xywh2xyxy(bbox)

    # Calculate the scaling ratio and padding
    ratio = min(resized_shape[0] / original_shape[0], resized_shape[1] / original_shape[1])
    new_unpad_shape = (int(round(original_shape[0] * ratio)), int(round(original_shape[1] * ratio)))
    dw = (resized_shape[1] - new_unpad_shape[1]) / 2  # x-padding
    dh = (resized_shape[0] - new_unpad_shape[0]) / 2  # y-padding

    # Remove padding from coordinates
    xyxy_coords[[0, 2]] -= (dw / resized_shape[1])
    xyxy_coords[[1, 3]] -= (dh /resized_shape[0])

    # Scale to original image size
    xyxy_coords[[0, 2]] = xyxy_coords[[0, 2]] *  resized_shape[1]/new_unpad_shape[1]
    xyxy_coords[[1, 3]] = xyxy_coords[[1, 3]] *  resized_shape[0]/new_unpad_shape[0]

    # Clip coordinates to be within the original image dimensions
    xyxy_coords[[0, 2]] = np.clip(xyxy_coords[[0, 2]], 0, 1)  
    xyxy_coords[[1, 3]] = np.clip(xyxy_coords[[1, 3]], 0, 1) 

    # Convert final xyxy to xywh (top-left corner)
    xywh_coords = xyxy2xywh(xyxy_coords)

    return xywh_coords



"""
Code to run Miew_ID and other re-identification models

(https://github.com/WildMeOrg/wbia-plugin-miew-id)

"""
from typing import Optional
from tqdm import tqdm
import pandas as pd
import numpy as np

import onnxruntime as ort

from animl.utils.general import get_onnx_device
from animl.generator import manifest_dataloader

MIEWID_SIZE = 440


def load_miew(file_path: str,
              device: Optional[str] = None):
    """
    Load MiewID from file path

    Args:
        file_path (str): file path to model file
        device (str): device to load model to

    Returns:
        loaded miewid model object
    """
    device = get_onnx_device(user_set=device)
    miew = ort.InferenceSession(file_path, providers=device)
    return miew


def extract_miew_embeddings(miew_model,
                            manifest: pd.DataFrame,
                            file_col: str = "filepath"):
    """
    Wrapper for MiewID embedding extraction

    Args:
        miew_model: MiewID model object
        manifest (pd.DataFrame): dataframe with columns 'filepath', 'emb_id'
        file_col (str): column name for file paths in manifest

    Returns:
        output (np.ndarray): array of extracted embeddings
    """
    if not {file_col}.issubset(manifest.columns):
        raise ValueError(f"DataFrame must contain '{file_col}' column.")

    output = []
    if isinstance(manifest, pd.DataFrame):

        dataloader = manifest_dataloader(manifest,
                                         file_col=file_col,
                                         crop=True,
                                         normalize={"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]},
                                         resize_width=MIEWID_SIZE,
                                         resize_height=MIEWID_SIZE,)
        for _, batch in tqdm(enumerate(dataloader), total=len(manifest)):
            img = batch[0]
            inp = miew_model.get_inputs()[0]
            outputs = miew_model.run(None, {inp.name: img})[0]
            output.extend(outputs)
        output = np.vstack(output)
    return output



"""
Generators and Dataloaders

Custom generators for training and inference

This version removes torch dependencies from the dataset and dataloader
so the module can be used without requiring PyTorch at runtime.
Images are returned as numpy arrays (C, H, W) with dtype float32 and
values scaled to [0, 1]. Batching is provided by a simple Python generator.
"""
import cv2
import numpy as np
from typing import Tuple, List, Optional, Iterable, Sequence
import pandas as pd
from pathlib import Path
from PIL import Image, ImageFile, ImageOps

from animl.file_management import IMAGE_EXTENSIONS, VIDEO_EXTENSIONS

ImageFile.LOAD_TRUNCATED_IMAGES = True


def Letterbox(resize_height: int,
              resize_width: int,
              image: Image.Image) -> Image.Image:
    # PIL image size: (width, height)
    width, height = image.size
    target_w, target_h = resize_width, resize_height

    # If aspect ratios are already the same (rounded to 2 decimals), just resize
    if round((width / height), 2) == round((target_w / target_h), 2):
        return image.resize((target_w, target_h), Image.BILINEAR)

    # Compute required padding to achieve target aspect ratio, do center pad
    target_ar = target_w / target_h
    src_ar = width / height

    if src_ar < target_ar:
        # source narrower -> pad left/right
        new_width = int(target_ar * height)
        pad_total = max(0, new_width - width)
        pad_left = pad_total // 2
        pad_right = pad_total - pad_left
        padded = ImageOps.expand(image, border=(pad_left, 0, pad_right, 0), fill=0)
    else:
        # source wider -> pad top/bottom
        new_height = int(width / target_ar)
        pad_total = max(0, new_height - height)
        pad_top = pad_total // 2
        pad_bottom = pad_total - pad_top
        padded = ImageOps.expand(image, border=(0, pad_top, 0, pad_bottom), fill=0)

    return padded.resize((target_w, target_h), Image.BILINEAR)


def Normalize(img: np.ndarray,
              mean: Sequence[float],
              std: Sequence[float],
              channel_axis: int = 0,
              scale: bool = False,
              out_dtype: Optional[np.dtype] = np.float32) -> np.ndarray:
    """
    Normalize image(s) with (img - mean) / std.

    img: HWC or CHW single image or batched NHWC/NCHW array.
    mean/std: length == n_channels (e.g., 3).
    channel_axis: axis index of channels, default -1 (H,W,C).
    scale: if True and img dtype is integer, divide by 255.0 first.
    out_dtype: output dtype (usually float32).
    """
    a = img.astype(out_dtype, copy=False)
    if scale and np.issubdtype(img.dtype, np.integer):
        a = a / 255.0

    mean = np.array(mean, dtype=out_dtype)
    std = np.array(std, dtype=out_dtype)

    # reshape mean/std to broadcast over image dims
    shape = [1] * a.ndim
    shape[channel_axis] = mean.size
    mean = mean.reshape(shape)
    std = std.reshape(shape)

    return (a - mean) / std


def pil_to_numpy_array(img: Image.Image) -> np.ndarray:
    """
    Convert PIL RGB image to numpy array with shape (C, H, W), dtype float32 scaled to [0,1].
    """
    arr = np.asarray(img, dtype=np.float32)  # H, W, C
    if arr.ndim == 2:  # grayscale -> replicate channels
        arr = np.stack([arr, arr, arr], axis=-1)
    # ensure RGB ordering; PIL.Image.convert("RGB") ensures this
    arr = arr.transpose(2, 0, 1)  # C, H, W
    arr = arr / 255.0
    return arr


class ManifestGenerator:
    '''
    Data generator that crops images on the fly, requires relative bbox coordinates,
    i.e. from MegaDetector.

    This class does NOT inherit from torch.utils.data.Dataset. It behaves like a
    sequence/iterable and can be indexed; it returns numpy arrays instead of torch.Tensors.
    '''
    def __init__(self, x: pd.DataFrame,
                 file_col: str = "filepath",
                 resize_height: int = 299,
                 resize_width: int = 299,
                 crop: bool = True,
                 crop_coord: str = 'relative',
                 normalize: bool = True,
                 letterbox: bool = False) -> None:
        self.x = x.reset_index(drop=True)
        self.file_col = file_col
        if self.file_col not in self.x.columns:
            raise ValueError(f"file_col '{self.file_col}' not found in dataframe columns")
        self.crop = crop
        if self.crop and not {'bbox_x', 'bbox_y', 'bbox_w', 'bbox_h'}.issubset(self.x.columns):
            raise ValueError("No bbox columns found for cropping")
        if crop_coord not in ['relative', 'absolute']:
            raise ValueError("crop_coord must be either 'relative' or 'absolute'")
        if 'frame' not in self.x.columns:
            self.x['frame'] = 0  # default frame 0 for images

        self.crop_coord = crop_coord
        self.resize_height = int(resize_height)
        self.resize_width = int(resize_width)
        self.buffer = 0
        self.normalize = normalize
        self.letterbox = bool(letterbox)

    def __len__(self) -> int:
        return len(self.x)

    def __getitem__(self, idx: int) -> Optional[Tuple[np.ndarray, str, int, np.ndarray]]:
        row = self.x.iloc[idx]
        filepath = row[self.file_col]
        frame = int(row.get('frame', 0))
        ext = Path(str(filepath)).suffix.lower()

        # read image or video frame
        if ext in VIDEO_EXTENSIONS:
            img = self.extract_frames(idx, filepath)
            if img is None:
                return None
        elif ext in IMAGE_EXTENSIONS:
            try:
                img = Image.open(filepath).convert('RGB')
            except OSError:
                print(f"Image {filepath} cannot be opened. Skipping.")
                return None
        else:
            print(f"File {filepath} is not a video or image. Skipping.")
            return None

        width, height = img.size

        # maintain aspect ratio if one dimension is zero (fallbacks)
        if self.resize_width > 0 and self.resize_height <= 0:
            self.resize_height = int(width / height * self.resize_width)
        elif self.resize_width <= 0 and self.resize_height > 0:
            self.resize_width = int(height / width * self.resize_height)

        # cropping if requested
        if self.crop:
            bbox_x = float(row['bbox_x'])
            bbox_y = float(row['bbox_y'])
            bbox_w = float(row['bbox_w'])
            bbox_h = float(row['bbox_h'])

            if self.crop_coord == 'relative':
                left = width * bbox_x
                top = height * bbox_y
                right = width * (bbox_x + bbox_w)
                bottom = height * (bbox_y + bbox_h)
            else:
                left = bbox_x
                top = bbox_y
                right = bbox_x + bbox_w
                bottom = bbox_y + bbox_h

            left = max(0, int(left) - self.buffer)
            top = max(0, int(top) - self.buffer)
            right = min(width, int(right) + self.buffer)
            bottom = min(height, int(bottom) + self.buffer)
            img = img.crop((left, top, right, bottom))

        # Letterbox
        if self.letterbox:
            img = Letterbox(self.resize_height, self.resize_width, img)
        else:
            img = img.resize((self.resize_width, self.resize_height), Image.BILINEAR)

        img_arr = pil_to_numpy_array(img)  # C,H,W
        img.close()

        # Normalize
        if isinstance(self.normalize, dict):
            img_arr = Normalize(img_arr,
                                mean=self.normalize.get("mean", [0.485, 0.456, 0.406]),
                                std=self.normalize.get("std", [0.229, 0.224, 0.225]))
        elif self.normalize is False:
            # unnormalize back to [0,255] if needed
            img_arr = img_arr * 255.0

        # return: img as numpy array (C,H,W), filepath str, frame int, shape np.array(height,width)
        return img_arr, str(filepath), int(frame), np.array((height, width), dtype=np.int32)

    def extract_frames(self, idx: int, filepath: str) -> Optional[Image.Image]:
        frame = int(self.x.loc[idx, 'frame'])
        cap = cv2.VideoCapture(str(filepath))
        if not cap.isOpened():  # corrupted video
            print(f"Video {filepath} cannot be opened. Skipping.")
            return None
        cap.set(cv2.CAP_PROP_POS_FRAMES, frame)
        ret, frame_img = cap.read()
        if not ret:
            print(f"Frame {frame} in video {filepath} cannot be read. Skipping.")
            cap.release()
            return None
        frame_img = cv2.cvtColor(frame_img, cv2.COLOR_BGR2RGB)
        img = Image.fromarray(frame_img)
        cap.release()
        cv2.destroyAllWindows()
        return img


def manifest_dataloader(manifest: pd.DataFrame,
                        file_col: str = "filepath",
                        crop: bool = True,
                        crop_coord: str = 'relative',
                        normalize: bool = True,
                        letterbox: bool = False,
                        resize_height: int = 299,
                        resize_width: int = 299) -> Iterable[Tuple[np.ndarray, List[str], np.ndarray, np.ndarray]]:
    '''
    Return a simple generator that yields batches of data as numpy arrays.

    Yields tuples: (batch_images, batch_filepaths, batch_frames, batch_hw)
      - batch_images: numpy array shape (B, C, H, W), dtype float32
      - batch_filepaths: list of file path strings length B
      - batch_frames: numpy array shape (B,) dtype int32
      - batch_hw: numpy array shape (B, 2) dtype int32 with (height, width)

    Notes:
      - This is a single-process generator. num_workers is intentionally ignored
        (no multiprocessing support) to avoid pulling in torch.
      - If a dataset item is None (skipped due to read errors), it will not be included.
    '''
    if crop is True and not any(manifest.columns.isin(["bbox_x"])):
        crop = False

    dataset_instance = ManifestGenerator(manifest,
                                         file_col=file_col,
                                         crop=crop,
                                         crop_coord=crop_coord,
                                         normalize=normalize,
                                         letterbox=letterbox,
                                         resize_width=resize_width,
                                         resize_height=resize_height)

    def batch_generator():
        for i in range(len(dataset_instance)):
            item = dataset_instance[i]
            if item is None:
                continue
            img_arr, path, frame, hw = item  # img_arr: C,H,W
            batch_np = np.stack([img_arr], axis=0)  # B,C,H,W
            yield batch_np, [path], np.array([frame], dtype=np.int32), np.stack([hw], axis=0)

    return batch_generator()



"""
Automated Pipeline Functions

@ Kyra Swanson 2023
"""
import os
import yaml
import pandas as pd
from pathlib import Path

from animl import (classification, detection, export, file_management,
                   video_processing, split)
from animl.utils import visualization


def from_paths(image_dir: str,
               detector_file: str,
               classifier_file: str,
               class_label: str = "class",
               sort: bool = False,
               visualize: bool = False,
               sequence: bool = False,
               detect_only: bool = False) -> pd.DataFrame:
    """
    This function is the main method to invoke all the sub functions
    to create a working directory for the image directory.

    Args:
        image_dir (str): directory path containing the images or videos.
        detector_file (str): file path of the MegaDetector model.
        classifier_file (str): file path of the classifier model.
        classlist_file (list): list of classes or species for classification.
        class_label: column in the class list that contains the label wanted
        batch_size (int): batch size for inference
        sort (bool): toggle option to create symlinks
        visualize (bool): if True, run visualization
        sequence (bool): if True, run sequence_classification

    Returns:
        pandas.DataFrame: Concatenated dataframe of animal and empty detections
    """
    # Create a working directory, build the file manifest from img_dir
    print("Searching directory...")
    working_dir = file_management.WorkingDirectory(image_dir)
    files = file_management.build_file_manifest(image_dir,
                                                out_file=working_dir.filemanifest,
                                                exif=True)
    # files["station"] = files["filepath"].apply(lambda x: x.split(os.sep)[-2])
    print(f"Found {len(files)} files.")

    # Obtain frames from videos
    all_frames = video_processing.extract_frames(files, frames=5, out_file=working_dir.imageframes)

    # Run detector
    print("Running images and video frames through detector...")
    if (file_management.check_file(working_dir.detections, output_type="Detections")):
        detections = file_management.load_data(working_dir.detections)
    else:
        detector = detection.load_detector(detector_file)
        md_results = detection.detect(detector,
                                      all_frames,
                                      resize_height=detection.MEGADETECTORv5_SIZE,
                                      resize_width=detection.MEGADETECTORv5_SIZE,
                                      checkpoint_path=working_dir.mdraw,
                                      checkpoint_frequency=1000)
        # Convert MD JSON to pandas dataframe, merge with manifest
        print("Converting MD JSON to dataframe and merging with manifest...")
        detections = detection.parse_detections(md_results, manifest=all_frames, out_file=working_dir.detections)

    if detect_only:
        print("Detection only flag set, skipping classification.")
        return detections

    # Extract animal detections from the rest
    animals = split.get_animals(detections)
    empty = split.get_empty(detections)

    # Use the classifier model to predict the species of animal detections
    print("Predicting species of animal detections...")
    classifier, class_list = classification.load_classifier(classifier_file)
    predictions_raw = classification.classify(classifier,
                                              animals,
                                              resize_height=classification.SDZWA_CLASSIFIER_SIZE,
                                              resize_width=classification.SDZWA_CLASSIFIER_SIZE,
                                              out_file=working_dir.predictions)
    if sequence:
        print("Classifying sequences...")
        manifest = classification.sequence_classification(animals, empty, predictions_raw,
                                                          class_list[class_label],
                                                          station_col='station',
                                                          empty_class="",
                                                          sort_columns=["station", "datetime", "frame"],
                                                          maxdiff=60)
    else:
        print("Classifying individual frames...")
        manifest = classification.single_classification(animals, empty, 
                                                        predictions_raw, 
                                                        class_list[class_label],
                                                        best=True)

    if sort:
        print("Sorting...")
        working_dir.activate_linkdir()
        manifest = export.export_folders(manifest, working_dir.linkdir)

    # Plot boxes
    if visualize:
        working_dir.activate_visdir()
        visualization.plot_all_bounding_boxes(manifest, working_dir.visdir, file_col='filepath', label_col='prediction')

    file_management.save_data(manifest, working_dir.results)
    print("Final Results in " + str(working_dir.results))

    return manifest


def from_config(config: str):
    """
    This function is the main method to invoke all the sub functions
    to create a working directory for the image directory.

    Args:
        config (str): path containing config file for inference

    Returns:
        pandas.DataFrame: Concatenated dataframe of animal and empty detections
    """
    print(f'Using config "{config}"')
    cfg = yaml.safe_load(open(config, 'r'))

    # get image dir and cuda defaults
    image_dir = cfg['image_dir']
    device = cfg.get('device', 'cpu')

    print("Searching directory...")
    # Create a working directory, default to image_dir
    working_dir = file_management.WorkingDirectory(cfg.get('working_dir', image_dir))
    files = file_management.build_file_manifest(image_dir,
                                                out_file=working_dir.filemanifest,
                                                exif=cfg.get('exif', True))
    print(f"Found {len(files)} files.")

    # Station Col
    station_dir = cfg.get('station_dir', None)
    if station_dir:
        files["station"] = files["filepath"].apply(lambda x: x.split(os.sep)[station_dir])

    # split out videos
    all_frames = video_processing.extract_frames(files, frames=5, out_file=working_dir.imageframes)

    print("Running images and video frames through detector...")
    if (file_management.check_file(working_dir.detections, output_type="Detections")):
        detections = file_management.load_data(working_dir.detections)
    else:
        detector = detection.load_detector(cfg['detector_file'])
        md_results = detection.detect(detector,
                                      all_frames,
                                      resize_height=detection.MEGADETECTORv5_SIZE,
                                      resize_width=detection.MEGADETECTORv5_SIZE,
                                      letterbox=cfg.get('letterbox', True),
                                      file_col=cfg.get('file_col_detection', 'filepath'),
                                      checkpoint_path=working_dir.mdraw,
                                      checkpoint_frequency=1000)
        # Convert MD JSON to pandas dataframe, merge with manifest
        print("Converting MD JSON to dataframe and merging with manifest...")
        detections = detection.parse_detections(md_results, manifest=all_frames, out_file=working_dir.detections)

    # Extract animal detections from the rest
    animals = split.get_animals(detections)
    empty = split.get_empty(detections)

    # Use the classifier model to predict the species of animal detections
    print("Predicting species...")
    # Load classifier
    classifier, class_list = classification.load_classifier(cfg['classifier_file'], cfg.get('class_list', None))

    predictions_raw = classification.classify(classifier,
                                              animals,
                                              resize_height=cfg.get('classifier_resize_height', classification.SDZWA_CLASSIFIER_SIZE),
                                              resize_width=cfg.get('classifier_resize_width', classification.SDZWA_CLASSIFIER_SIZE),
                                              file_col=cfg.get('file_col_classification', 'filepath'),
                                              out_file=working_dir.predictions)

    # Convert predictions to labels
    if station_dir:
        manifest = classification.sequence_classification(animals, empty, predictions_raw,
                                                          class_list[cfg.get('class_label_col', 'class')],
                                                          station_col='station',
                                                          empty_class=cfg['empty_class'],
                                                          sort_columns=["station", "datetime", "frame"],
                                                          file_col=cfg.get('file_col_classification', 'frame'),
                                                          maxdiff=60)
    else:
        manifest = classification.single_classification(animals, empty, 
                                                        predictions_raw, 
                                                        class_list[cfg.get('class_label_col', 'class')],
                                                        file_col=cfg.get('file_col_classification', 'filepath'),
                                                        best=cfg.get('best_only', True))

    if cfg.get('sort', True):
        print("Sorting...")
        working_dir.activate_linkdir()
        manifest = export.export_folders(manifest, working_dir.linkdir, copy=cfg.get('copy', False))

    # Plot boxes
    if cfg.get('visualize', False):
        working_dir.activate_visdir()
        visualization.plot_all_bounding_boxes(manifest, working_dir.visdir, file_col='filepath', label_col='prediction')

    file_management.save_data(manifest, working_dir.results)
    print("Final Results in " + str(working_dir.results))

    return manifest



'''
Tools for Saving, Loading, and Using Species Classifiers

@ Kyra Swanson 2023
'''
import json
from typing import Optional, Union
import pandas as pd
import numpy as np
from pathlib import Path
from time import time
from tqdm import tqdm
import onnxruntime

from animl import generator, file_management
from animl.utils.general import get_onnx_device, softmax


SDZWA_CLASSIFIER_SIZE = 299


def load_classifier(model_path: str,
                    classes: Union[int, str, Path, pd.DataFrame] = None):
    '''
    Creates a model instance and loads the latest model state weights.

    Args:
        model_path (str): file or directory path to model weights
        classes (int | str | Path | pd.DataFrame): number of classes, path to associated class list, or pd.DataFrame of class list
        device (str): specify to run on cpu or gpu
        architecture (str): expected model architecture

    Returns:
        model: model object of given architecture with loaded weights
        class_list: list of class names
    '''
    class_list = None
    model_path = Path(model_path)
    # check to make sure GPU is available if chosen
    providers = get_onnx_device()

    print(f'Loading model at {model_path}')
    model = onnxruntime.InferenceSession(model_path, providers=providers)
    model.framework = "onnx"

    # get number of classes
    if isinstance(classes, str) or isinstance(classes, Path):
        class_list = load_class_list(classes)
    elif isinstance(classes, pd.DataFrame):
        class_list = classes
    else:
        class_list = None
    # try to load class dict from metadata
    props = model.get_modelmeta().custom_metadata_map
    if "class_dict" in props:
        print("Loaded class_dict from ONNX metadata.")
        class_dict = json.loads(props["class_dict"])
        class_list = [class_dict[str(i)] for i in range(len(class_dict))]
        class_list = pd.DataFrame({'class': class_list})

        # no need to return epoch
    return model, class_list


def load_class_list(classlist_file):
    """
    Return classlist file as pd.DataFrame.

    Args:
        classlist_file (str): file path to class list

    Returns:
        pd.DataFrame of class list
    """
    if not Path(classlist_file).is_file():
        raise FileNotFoundError(f"Class list file not found at {classlist_file}")
    return pd.read_csv(classlist_file)


def classify(model,
             detections,
             resize_width: int = 480,
             resize_height: int = 480,
             file_col: str = 'filepath',
             crop: bool = True,
             normalize: bool = True,
             out_file: Optional[str] = None):
    """
    TODO: align with R version
    Predict species using classifier model.

    Args:
        model: preloaded classifier model
        detections (mult): dataframe of (animal) detections, list of filepaths or filepath str
        resize_width (int): image width input size
        resize_height (int): image height input size
        file_col (str): column name containing file paths
        crop (bool): use bbox to crop images before feeding into model
        normalize (bool): normalize the tensor before inference
        batch_size (int): data generator batch size
        num_workers (int): number of cores
        device (str): specify to run model on cpu or gpu, default to cpu
        out_file (str): path to save prediction results to

    Returns:
        detections (pd.DataFrame): MD detections with classifier prediction and confidence
    """
    if file_management.check_file(out_file, output_type="Classification results"):
        return file_management.load_data(out_file).to_numpy()

    # initialize lists
    raw_output = []

    # Manifest
    if isinstance(detections, pd.DataFrame):
        if file_col not in detections.columns:
            raise ValueError(f"file_col {file_col} not found in manifest columns")
        # no frame column, assume all images and set to 0
        if 'frame' not in detections.columns:
            print("Warning: 'frame' column not found in manifest columns. Defaulting to 0 assuming images.")
            detections['frame'] = 0

        dataset = generator.manifest_dataloader(detections, file_col=file_col, crop=crop,
                                                resize_width=resize_width, resize_height=resize_height,
                                                normalize=normalize)
    # Single File
    elif isinstance(detections, str):
        detections = pd.DataFrame({file_col: detections, 'frame': 0}, index=[0])
        dataset = generator.manifest_dataloader(detections, file_col=file_col, crop=False,
                                                resize_width=resize_width, resize_height=resize_height,
                                                normalize=normalize)
    # List of Files
    elif isinstance(detections, list):
        detections = pd.DataFrame({file_col: detections, 'frame': 0}, index=range(len(detections)))
        dataset = generator.manifest_dataloader(detections, file_col=file_col, crop=False,
                                                resize_width=resize_width, resize_height=resize_height,
                                                normalize=normalize)
    else:
        raise AssertionError("Input must be a data frame of crops, single file path or vector of file paths.")

    # Predict
    start_time = time()
    for _, batch in tqdm(enumerate(dataset), total=len(detections)):
        image = batch[0]
        output = model.run(None, {model.get_inputs()[0].name: image})[0]
        raw_output.extend(softmax(output))

    raw_output = np.vstack(raw_output)

    if out_file:
        file_management.save_data(pd.DataFrame(raw_output), out_file)

    print(f"\nFinished classification. Total images processed: {len(raw_output)} at {round(len(raw_output)/(time() - start_time), 1)} img/s.")

    return raw_output


def single_classification(animals: pd.DataFrame,
                          empty: Optional[pd.DataFrame],
                          predictions_raw: np.array,
                          class_list: list,
                          best: bool = False):
    """
    Get maximum likelihood prediction from softmaxed logits.

    Args:
        animals (pd.DataFrame): animal detections from manifest
        empty (Optional) (pd.DataFrame): empty detections from manifest
        predictions_raw (np.array): softmaxed logits from classify()
        class_list (list): class list associated with model
        best (bool): whether to return one prediction per file

    Returns:
        animals dataframe with "prediction" label an "confidence" columns
    """
    class_list = pd.Series(class_list)

    # convert None to empty dataframe fo concat
    if empty is None:
        empty = pd.DataFrame()

    if not animals.empty:
        animals = animals.reset_index(drop=True)
        animals["prediction"] = class_list.values[np.argmax(predictions_raw, axis=1)]
        animals["confidence"] = animals["conf"].mul(np.max(predictions_raw, axis=1))

    manifest = pd.concat([animals if not animals.empty else None, empty if not empty.empty else None]).reset_index(drop=True)

    # remove empties from videos
    files = manifest.groupby('filepath')
    for f, file in files:
        if file['extension'].iloc[0] in file_management.VIDEO_EXTENSIONS:
            predictions = file['prediction'].unique()
            if 'empty' in predictions and len(predictions) > 1:
                real_prediction = predictions[predictions != 'empty'][0]
                manifest.loc[manifest['filepath'] == f, 'prediction'] = real_prediction

    # best guess
    if best:
        # take most confident guess    
        manifest = manifest.sort_values("confidence", ascending=False)
        manifest = manifest.drop_duplicates(subset="filepath", keep="first")
    
    return manifest.reset_index(drop=True)


def sequence_classification(animals: pd.DataFrame,
                            empty: Optional[pd.DataFrame],
                            predictions_raw: np.array,
                            class_list: pd.DataFrame,
                            station_col: str,
                            empty_class: str = "",
                            sort_columns: list[str] = None,
                            file_col: str = "filepath",
                            maxdiff: int = 60):
    """
    Applies class labels to images based on sequential information.

    This function applies image classifications at a sequence level by leveraging information from
    multiple images. A sequence is defined as all images at the same camera and station where the
    time between consecutive images is <=maxdiff. This can improve classification accuracy, but
    assumes that only one species is present in each sequence. If you regularly expect multiple
    species to occur in an image or sequence don't use this function.

    Args:
        animals (pd.DataFrame): Sub-selection of all images that contain animals
        empty (Optional) (pd.DataFrame): Sub-selection of all images that do not contain animals
        predictions_raw (Numpy Array of Numpy Arrays): Logits of all entries in "animals"
        class_list (pd.DataFrame): class list associated with classifier model
        station_col (str): The name of the station column
        empty_class (str) (Optional): the name of class_list 'empty' label
        sort_columns (List of Strings): Defines sorting order for the DataFrame
        file_col (str): The name of the filepath column
        maxdiff (int): Maximum time difference in seconds between any two images in a sequence

    Returns:
        final_df (pd.DataFrame): Sequence classified data from both animals and empty

    Raises:
        Exception: If 'sort_columns' is not a list or is empty
        Exception: If 'station_col' is not a string or is empty
        Exception: If 'empty' is defined and is not a pandas DataFrame
        Exception: If maxdiff is defined and is not a positive number
    """
    if not isinstance(station_col, str) or station_col == '':
        raise Exception("'station_col' must be a non-empty string")

    # Sanity check to verify that empty is a Pandas DataFrame, if defined
    if empty is not None and not isinstance(animals, pd.DataFrame):
        raise Exception("'empty' must be a DataFrame")

    # Sanity check to verify that maxdiff is a positive number
    if not isinstance(maxdiff, (int, float)) or maxdiff < 0:
        raise Exception("'maxdiff' must be a number >= 0")

    if not {file_col}.issubset(animals.columns):
        raise ValueError(f"DataFrame must contain '{file_col}' column.")

    if not {"datetime"}.issubset(animals.columns):
        raise ValueError("DataFrame must contain 'datetime' column.")

    if "conf" not in animals.columns:
        animals["conf"] = 1

    if empty_class > "":
        empty_col = class_list[class_list == empty_class].index[0]
    else:
        empty_col = None

    if empty is not None or not empty.empty:
        empty["ID"] = range(0, empty.shape[0])
        predempty = empty.pivot(index="ID", columns="prediction", values="confidence")
        # Replace NaN with 0
        predempty = predempty.fillna(0)
        predempty = pd.concat([pd.DataFrame(np.zeros((empty.shape[0], len(class_list)))), predempty], axis=1)

        if empty_class > "":
            if 'empty' in predempty.columns:
                predempty[empty_col] = predempty["empty"]
                predempty = predempty.drop("empty", axis=1)
            class_list = pd.concat([class_list,
                                    pd.Series([x for x in empty["prediction"].unique() if x != "empty"])], ignore_index=True)

        else:
            class_list = pd.concat([class_list, pd.Series(empty["prediction"].unique())], ignore_index=True)
            empty_col = predempty.columns.get_loc("empty")

        # placeholders
        animals["prediction"] = list(class_list[np.argmax(predictions_raw, axis=1)])
        animals["confidence"] = animals["conf"].mul(np.max(predictions_raw, axis=1))

        empty["conf"] = 1
        animals_merged = pd.concat([animals, empty.iloc[:, :-1]]).reset_index(drop=True)  # dont add ID column
        predictions = np.hstack((predictions_raw,
                                 np.zeros((predictions_raw.shape[0], len(predempty.columns) - predictions_raw.shape[1]))))
        # concat
        predictions = np.vstack((predictions, np.array(predempty)))

    if sort_columns is None:
        sort_columns = [station_col, "datetime"]

    animals_merged['datetime'] = pd.to_datetime(animals_merged['datetime'], format="%Y-%m-%d %H:%M:%S")

    sort = animals_merged.sort_values(by=sort_columns).index
    animals_sort = animals_merged.loc[sort].reset_index(drop=True)
    predsort = predictions[sort]

    conf_placeholder = np.zeros(len(animals_sort))
    predict_placeholder = np.empty(len(animals_sort), dtype='U30')
    sequence_placeholder = np.zeros(len(animals_sort))

    i = 0
    s = 0
    while i < len(animals_sort):
        rows = [i]
        last_index = i+1

        while (last_index < len(animals_sort) and not pd.isna(animals_sort.loc[i, "datetime"]) and
               not pd.isna(animals_sort.loc[last_index, "datetime"]) and
               animals_sort.loc[last_index, station_col] == animals_sort.loc[i, station_col] and
               (animals_sort.loc[last_index, "datetime"] - animals_sort.loc[i, "datetime"]).total_seconds() <= maxdiff):
            rows.append(last_index)
            last_index += 1

        rows = np.array(rows)

        # multiple detections in sequence
        if len(rows) > 1:
            predclass = np.argmax(predsort[rows], axis=1)

            # no empties
            if empty_col is None or empty_col not in predclass:
                predsort_confidence = predsort[rows] * np.reshape(animals_sort.loc[rows, 'conf'].values, (-1, 1))
                predbest = np.mean(predsort_confidence, axis=0)
                conf_placeholder[rows] = np.max(predsort_confidence[:, np.argmax(predbest)])
                predict_placeholder[rows] = class_list[np.argmax(predbest)]
                sequence_placeholder[rows] = int(s)

            else:
                mask = pd.DataFrame((predclass == empty_col))
                filtered_animals = animals_sort.loc[rows, file_col].reset_index(drop=True)
                sum_values = mask.groupby(filtered_animals).sum()
                count_values = mask.groupby(filtered_animals).count()

                sel_all_empty = count_values == sum_values
                sel_mixed = np.where(filtered_animals.isin(sel_all_empty[~sel_all_empty[0]].index))[0]
                sel_no_empties = np.where(filtered_animals.isin(sel_all_empty[~sel_all_empty[0]].index) & (predclass != empty_col))[0]

                if len(sel_mixed) > 0 and len(sel_no_empties) > 0:
                    predsort_confidence = predsort[rows[sel_no_empties]] * np.reshape(animals_sort.loc[rows[sel_no_empties], 'conf'].values, (-1, 1))
                    predbest = np.mean(predsort_confidence, axis=0)
                    conf_placeholder[rows[sel_mixed]] = np.max(predsort_confidence[:, np.argmax(predbest)])
                    predict_placeholder[rows[sel_mixed]] = class_list[np.argmax(predbest)]
                    sequence_placeholder[rows[sel_mixed]] = int(s)
                for file in sel_all_empty[sel_all_empty[0]].index:
                    empty_row = np.where(animals_sort[file_col] == file)
                    predsort_confidence = predsort[empty_row] * np.reshape(animals_sort.loc[empty_row, 'conf'].values, (-1, 1))
                    predbest = np.mean(predsort_confidence, axis=0)
                    conf_placeholder[empty_row] = np.max(predsort_confidence[:, np.argmax(predbest)])
                    sequence_placeholder[empty_row] = int(s)
                    predict_placeholder[empty_row] = class_list[np.argmax(predbest)]
        # single row in sequence
        else:
            predbest = predsort[rows]
            conf_placeholder[rows] = np.max(predbest * animals_sort.loc[rows, 'conf'].values)
            predict_placeholder[rows] = class_list[np.argmax(predbest)]
            sequence_placeholder[rows] = int(s)

        i = last_index
        s += 1

    animals_sort['confidence'] = conf_placeholder
    animals_sort['prediction'] = predict_placeholder
    animals_sort['sequence'] = sequence_placeholder

    return animals_sort



from animl.utils import general
from animl.utils import visualization

from animl.utils.general import (clip_coords, convert_minxywh_to_absxyxy,
                                 get_onnx_device, letterbox, normalize_boxes,
                                 scale_letterbox, softmax, tensor_to_onnx,
                                 xyn2xy, xywh2xyxy, xywhc2xyxy, xywhn2xyxy,
                                 xyxy2xywh, xyxyc2xywh, xyxyc2xywhn,)
from animl.utils.visualization import (MD_COLORS, MD_LABELS,
                                       plot_all_bounding_boxes, plot_box,
                                       plot_from_file,)

__all__ = ['MD_COLORS', 'MD_LABELS', 'clip_coords',
           'convert_minxywh_to_absxyxy', 'general', 'get_onnx_device',
           'letterbox', 'normalize_boxes', 'plot_all_bounding_boxes',
           'plot_box', 'plot_from_file', 'scale_letterbox', 'softmax',
           'tensor_to_onnx', 'visualization', 'xyn2xy', 'xywh2xyxy',
           'xywhc2xyxy', 'xywhn2xyxy', 'xyxy2xywh', 'xyxyc2xywh',
           'xyxyc2xywhn']



from animl.reid import distance
from animl.reid import inference

from animl.reid.distance import (compute_batched_distance_matrix,
                                 compute_distance_matrix, cosine_distance,
                                 euclidean_squared_distance, remove_diagonal,)
from animl.reid.inference import (MIEWID_SIZE, extract_miew_embeddings,
                                  load_miew,)

__all__ = ['MIEWID_SIZE', 'compute_batched_distance_matrix',
           'compute_distance_matrix', 'cosine_distance', 'distance',
           'euclidean_squared_distance', 'extract_miew_embeddings',
           'inference', 'load_miew', 'remove_diagonal']



__version__ = '3.2.0'

from animl import classification
from animl import detection
from animl import export
from animl import file_management
from animl import generator
from animl import pipeline
from animl import reid
from animl import split
from animl import utils
from animl import video_processing

from animl.classification import (SDZWA_CLASSIFIER_SIZE, classify,
                                  load_class_list, load_classifier,
                                  sequence_classification,
                                  single_classification,)
from animl.detection import (MEGADETECTORv5_SIZE, convert_onnx_detections,
                             detect, load_detector, parse_detections,)
from animl.export import (export_camtrapR, export_coco, export_folders,
                          export_megadetector, export_timelapse, remove_link,
                          update_labels_from_folders,)
from animl.file_management import (IMAGE_EXTENSIONS, VALID_EXTENSIONS,
                                   VIDEO_EXTENSIONS, WorkingDirectory,
                                   active_times, build_file_manifest,
                                   check_file, load_data, load_json, save_data,
                                   save_detection_checkpoint, save_json,
                                   sequence_calculation,)
from animl.generator import (Letterbox, ManifestGenerator, Normalize,
                             manifest_dataloader, pil_to_numpy_array,)
from animl.pipeline import (from_config, from_paths,)
from animl.reid import (MIEWID_SIZE, compute_batched_distance_matrix,
                        compute_distance_matrix, cosine_distance, distance,
                        euclidean_squared_distance, extract_miew_embeddings,
                        inference, load_miew, remove_diagonal,)
from animl.split import (get_animals, get_empty,)
from animl.utils import (MD_COLORS, MD_LABELS, clip_coords,
                         convert_minxywh_to_absxyxy, general, get_onnx_device,
                         letterbox, normalize_boxes, plot_all_bounding_boxes,
                         plot_box, plot_from_file, scale_letterbox, softmax,
                         tensor_to_onnx, visualization, xyn2xy, xywh2xyxy,
                         xywhc2xyxy, xywhn2xyxy, xyxy2xywh, xyxyc2xywh,
                         xyxyc2xywhn,)
from animl.video_processing import (count_frames, extract_frames,
                                    get_frame_as_image,)

__all__ = ['IMAGE_EXTENSIONS', 'Letterbox', 'MD_COLORS', 'MD_LABELS',
           'MEGADETECTORv5_SIZE', 'MIEWID_SIZE', 'ManifestGenerator',
           'Normalize', 'SDZWA_CLASSIFIER_SIZE', 'VALID_EXTENSIONS',
           'VIDEO_EXTENSIONS', 'WorkingDirectory', 'active_times',
           'build_file_manifest', 'check_file', 'classification', 'classify',
           'clip_coords', 'compute_batched_distance_matrix',
           'compute_distance_matrix', 'convert_minxywh_to_absxyxy',
           'convert_onnx_detections', 'cosine_distance', 'count_frames',
           'detect', 'detection', 'distance', 'euclidean_squared_distance',
           'export', 'export_camtrapR', 'export_coco', 'export_folders',
           'export_megadetector', 'export_timelapse', 'extract_frames',
           'extract_miew_embeddings', 'file_management', 'from_config',
           'from_paths', 'general', 'generator', 'get_animals', 'get_empty',
           'get_frame_as_image', 'get_onnx_device', 'inference', 'letterbox',
           'load_class_list', 'load_classifier', 'load_data', 'load_detector',
           'load_json', 'load_miew', 'manifest_dataloader', 'normalize_boxes',
           'parse_detections', 'pil_to_numpy_array', 'pipeline',
           'plot_all_bounding_boxes', 'plot_box', 'plot_from_file', 'reid',
           'remove_diagonal', 'remove_link', 'save_data',
           'save_detection_checkpoint', 'save_json', 'scale_letterbox',
           'sequence_calculation', 'sequence_classification',
           'single_classification', 'softmax', 'split', 'tensor_to_onnx',
           'update_labels_from_folders', 'utils', 'video_processing',
           'visualization', 'xyn2xy', 'xywh2xyxy', 'xywhc2xyxy', 'xywhn2xyxy',
           'xyxy2xywh', 'xyxyc2xywh', 'xyxyc2xywhn']



"""
Object Detection Module

Functions for loading MegaDetector, as well as custom YOLO models
parse_detections() converts json output into a dataframe

"""
import argparse
from typing import Optional
import time
import numpy as np
import pandas as pd
from pathlib import Path
from tqdm import tqdm
import onnxruntime as ort

from animl import file_management
from animl.generator import manifest_dataloader
from animl.utils.general import normalize_boxes, xyxy2xywh, scale_letterbox, get_onnx_device

MEGADETECTORv5_SIZE = 1280


def load_detector(model_path: str):
    """
    Load Detector model from filepath.

    Args:
        model_path (str): path to model file
        model_type (str): type of model expected ["MDV5", "MDV6", "YOLO", "ONNX"]
        device (str): specify to run on cpu or gpu

    Returns:
        object: loaded model object
    """
    if not Path(model_path).is_file():
        raise FileNotFoundError(f"Model file not found at {model_path}")

    providers = get_onnx_device()
    model = ort.InferenceSession(model_path, providers=providers)
    model.model_type = "onnx"
    return model


def detect(detector,
           image_file_names,
           resize_width: int,
           resize_height: int,
           letterbox: bool = True,
           confidence_threshold: float = 0.1,
           file_col: str = 'filepath',
           checkpoint_path: Optional[str] = None,
           checkpoint_frequency: int = -1) -> list[dict]:
    """
    Runs Detector model on a batches of image files.

    Args:
        detector (object): preloaded detector model
        image_file_names (mult): list of image filenames, a single image filename, or manifest
                                    containing a list of images.
        resize_width (int): width to resize images to
        resize_height (int): height to resize images to
        letterbox (bool): if True, resize and pad image to keep aspect ratio, else resize without padding
        confidence_threshold (float): only detections above this threshold are returned
        file_col (str): column name containing file paths
        device (str): specify to run on cpu or gpu
        checkpoint_path (str): path to checkpoint file
        checkpoint_frequency (int): write results to checkpoint file every N images

    Returns:
        list: list of dicts, each dict represents detections on one image
    """
    # Single image filepath
    if isinstance(image_file_names, str):
        # convert img path to tensor
        batch_from_dataloader = manifest_dataloader(pd.DataFrame([[image_file_names, 0]], columns=['filepath', 'frame']),
                                                    crop=False,
                                                    normalize=True,
                                                    letterbox=letterbox,
                                                    resize_width=resize_width,
                                                    resize_height=resize_height)
        image_tensors = batch_from_dataloader[0]  # Tensor of images for the current batch
        current_image_paths = batch_from_dataloader[1]  # List of image names for the current batch
        image_sizes = batch_from_dataloader[2]  # List of original image sizes for the current batch

        input_name = detector.get_inputs()[0].name
        outputs = detector.run(None, {input_name: image_tensors.cpu().numpy()})[0]
        results = convert_onnx_detections(outputs, image_tensors, current_image_paths, image_sizes, letterbox)
        return results

    # list of image filepaths
    elif isinstance(image_file_names, list):
        # create a data frame from list of image paths
        manifest = pd.DataFrame(image_file_names, columns=[file_col])
        # no frame column, assume all images and set to 0
        manifest['frame'] = 0

    # full manifest, select file_col
    elif isinstance(image_file_names, pd.DataFrame):
        if file_col not in image_file_names.columns:
            raise ValueError(f"file_col {file_col} not found in manifest columns")
        # no frame column, assume all images and set to 0
        if 'frame' not in image_file_names.columns:
            print("Warning: 'frame' column not found in manifest columns. Defaulting to 0 assuming images.")
            image_file_names['frame'] = 0
        # create a list of image paths
        manifest = image_file_names[[file_col, 'frame']]

    # single row pd.Series, select file_col
    elif isinstance(image_file_names, pd.Series):
        if file_col not in image_file_names.index:
            raise ValueError(f"file_col {file_col} not found in Series index")
        if 'frame' not in image_file_names.index:
            print("Warning: 'frame' column not found in Series index. Defaulting to 0 assuming images.")
            image_file_names['frame'] = 0
        # create a data frame from image_file_names
        manifest = pd.DataFrame(image_file_names).T
    # column from pd.DataFrame, expected input
    else:
        raise ValueError('image_file_names is not a recognized object')

    # load checkpoint
    if checkpoint_path and file_management.check_file(checkpoint_path, output_type="Megadetector raw output"):
        results = file_management.load_json(checkpoint_path).get('images')
        already_processed = set([r['filepath'] for r in results])
        manifest = image_file_names[~image_file_names[file_col].isin(already_processed)][[file_col, 'frame']].reset_index(drop=True)
        if manifest.empty:
            print("All images have already been processed. Exiting.")
            return results
    else:
        results = []
        image_file_names = set(image_file_names)

    count = 0

    # create dataloader
    dataloader = manifest_dataloader(manifest,
                                     crop=False,
                                     normalize=True,
                                     letterbox=letterbox,
                                     resize_width=resize_width,
                                     resize_height=resize_height)

    start_time = time.time()
    for _, batch in tqdm(enumerate(dataloader), total=len(manifest)):
        count += 1

        image_tensors = batch[0]  # Tensor of images for the current batch
        current_image_paths = batch[1]  # List of image names for the current batch
        current_frames = batch[2]  # List of frame numbers for the current batch
        current_sizes = batch[3]  # List of original image sizes for the current batch

        # ONNX Runtime inference
        input_name = detector.get_inputs()[0].name
        outputs = detector.run(None, {input_name: image_tensors})[0]
        outputs = convert_onnx_detections(outputs, confidence_threshold,
                                          image_tensors, current_image_paths, current_frames, current_sizes, letterbox)
        # Process outputs to match expected format
        results.extend(outputs)

        # Write a checkpoint if necessary
        if checkpoint_frequency != -1 and count % checkpoint_frequency == 0:
            print('Writing a new checkpoint after having processed {} images since last restart'.format(count))
            file_management.save_detection_checkpoint(checkpoint_path, results)

    print(f"\nFinished detection. Total images processed: {len(results)} at {round(len(results)/(time.time() - start_time), 1)} img/s.")
    if checkpoint_path:
        file_management.save_detection_checkpoint(checkpoint_path, results)

    return results


def convert_onnx_detections(predictions: list,
                            confidence_threshold: float,
                            image_tensors: list,
                            image_paths: list,
                            image_frames: list,
                            image_sizes: list,
                            letterbox: bool) -> pd.DataFrame:
    # Process ONNX predictions
    results = []

    for i, pred in enumerate(predictions):

        boxes = pred[:, :4]  # Bounding box coordinates
        conf = pred[:, 4]  # Confidence scores
        category = pred[:, 5]  # Class labels as integers
        max_detection_conf = conf.max() if len(conf) > 0 else 0

        if len(conf) == 0 or all(c < confidence_threshold for c in conf):
            data = {'filepath': str(image_paths[i]),
                    'frame': int(image_frames[i]),
                    'max_detection_conf': float(round(max_detection_conf, 4)),
                    'detections': []}
            results.append(data)
        else:
            detections = []
            for j in range(len(conf)):
                if conf[j] < confidence_threshold:
                    continue

                bbox = normalize_boxes(boxes[j], image_tensors[i].shape[1:])
                bbox = xyxy2xywh(bbox)
                if bbox.all() == 0:
                    continue

                if letterbox:
                    bbox = scale_letterbox(bbox, image_tensors[i].shape[1:], image_sizes[i, :])

                detection = {
                    'bbox_x': float(round(bbox[0], 4)),
                    'bbox_y': float(round(bbox[1], 4)),
                    'bbox_w': float(round(bbox[2], 4)),
                    'bbox_h': float(round(bbox[3], 4)),
                    'conf': float(round(conf[j], 4)),
                    'category': int(category[j]+1)
                }
                detections.append(detection)
            data = {'filepath': str(image_paths[i]),
                    'frame': int(image_frames[i]),
                    'max_detection_conf': float(round(max_detection_conf, 4)),
                    'detections': detections}
            results.append(data)

    return results


def parse_detections(results: list,
                     manifest: Optional[pd.DataFrame] = None,
                     out_file: Optional[str] = None,
                     threshold: float = 0,
                     file_col: str = "filepath"):
    """
    Converts listed output from detector to DataFrame.

    Args:
        results (list): md output dicts
        manifest (pd.DataFrame): full file manifest, if not None, merge md predictions automatically
        out_file (str): path to save dataframe
        threshold (float): parse only detections above given confidence threshold
        file_col (str): if manifest, merge results onto file_col

    Returns:
        df (pd.DataFrame): formatted md outputs, one row per detection
    """
    # load checkpoint
    if file_management.check_file(out_file, output_type="Detections"):  # checkpoint comes back empty
        df = file_management.load_data(out_file)
        already_processed = set([row['filepath'] for row in df])

    else:
        df = pd.DataFrame(columns=('filepath', 'frame', 'max_detection_conf', 'category', 'conf',
                                   'bbox_x', 'bbox_y', 'bbox_w', 'bbox_h'))
        already_processed = set()

    if not isinstance(results, list):
        raise AssertionError("MD results input must be list")

    if len(results) == 0:
        raise AssertionError("'results' contains no detections")

    lst = []

    for frame in results:
        # pass if already analyzed
        if frame['filepath'] in already_processed:
            continue

        try:
            detections = frame['detections']
        except KeyError:
            print('File error ', frame['filepath'])
            continue

        if len(detections) == 0:
            data = {'filepath': frame['filepath'],
                    'frame': frame['frame'],
                    'max_detection_conf': frame['max_detection_conf'],
                    'category': 0, 'conf': None, 'bbox_x': None,
                    'bbox_y': None, 'bbox_w': None, 'bbox_h': None}
            lst.append(data)

        else:
            for detection in detections:
                if (detection['conf'] > threshold):
                    data = {'filepath': frame['filepath'],
                            'frame': frame['frame'],
                            'max_detection_conf': frame['max_detection_conf'],
                            'category': detection['category'], 'conf': detection['conf'],
                            'bbox_x': np.clip(detection['bbox_x'], 0, 1),
                            'bbox_y': np.clip(detection['bbox_y'], 0, 1),
                            'bbox_w': np.clip(detection['bbox_w'], 0, 1),
                            'bbox_h': np.clip(detection['bbox_h'], 0, 1)}
                    lst.append(data)

    df = pd.DataFrame(lst)

    if manifest is not None:
        if file_col in manifest.columns:
            df = manifest.merge(df, left_on=[file_col, 'frame'], right_on=["filepath", "frame"], how='left')
        else:
            raise ValueError("Please provide a manifest with a valid file_col to merge results onto.")

    if out_file:
        file_management.save_data(df, out_file)

    return df


"""
File Management Module

This module provides functions and classes for managing files and directories.

@ Kyra Swanson 2023
"""
import json
from shutil import copyfile
from pathlib import Path, PosixPath
from datetime import datetime, timedelta
import pandas as pd
import numpy as np
import PIL
import cv2
from typing import Optional


IMAGE_EXTENSIONS = {'.png', '.jpg', '.jpeg', ".tiff", '.tif"'}
VIDEO_EXTENSIONS = {".mp4", ".avi", ".mov", ".wmv",
                    ".mpg", ".mpeg", ".asf", ".m4v"}
VALID_EXTENSIONS = IMAGE_EXTENSIONS | VIDEO_EXTENSIONS


def build_file_manifest(image_dir: str,
                        exif: bool = True,
                        out_file: Optional[str] = None,
                        offset: int = 0,
                        recursive: bool = True):
    """
    Find Image/Video Files and Gather exif Data.

    Args:
        image_dir (str): directory of files to analyze
        exif (bool): returns date and time info from exif data, defaults to True
        out_file (str): file path to which the dataframe should be saved
        offset (int): add timezone offset in hours to datetime column
        recursive (bool): recursively search through all child directories

    Returns:
        files (pd.DataFrame): list of files with or without file modify dates
    """
    image_dir = Path(image_dir)
    if check_file(out_file, output_type="Manifest"):
        return load_data(out_file)
    if not image_dir.is_dir():
        raise FileNotFoundError(f"The given directory: {image_dir}, does not exist.")

    files = Path(image_dir).rglob("*.*") if recursive else Path(image_dir).glob("*.*")

    # only keep images and videos
    files = [str(f) for f in files if Path(f).suffix.lower() in VALID_EXTENSIONS]

    # no files found, return empty dataframe
    if not files:
        return pd.DataFrame()

    files = pd.DataFrame(files, columns=["filepath"])
    files["filename"] = files["filepath"].apply(lambda x: Path(x).name)
    files["extension"] = files["filepath"].apply(lambda x: Path(x).suffix.lower())

    invalid = []

    def check_time(timestamp):
        input_formats = ['%Y:%m:%d %H:%M:%S', "%d-%m-%Y %H:%M", "%Y/%m/%d %H:%M:%S"]
        desired_format = '%Y-%m-%d %H:%M:%S'
        try:
            # If it already matches, return as is
            if datetime.strptime(timestamp, desired_format).strftime(desired_format) == timestamp:
                return timestamp
        except ValueError:
            pass
        # Try other input formats
        for fmt in input_formats:
            try:
                newtimestamp = datetime.strptime(timestamp, fmt)
                return newtimestamp.strftime(desired_format)
            except ValueError:
                continue
        # timestamp not recognized
        return None

    if exif:
        for i, row in files.iterrows():
            if row["extension"] in IMAGE_EXTENSIONS:
                try:
                    img = PIL.Image.open(row['filepath'])
                    files.loc[i, "width"] = img.size[0]
                    files.loc[i, "height"] = img.size[1]
                    files.loc[i, "createdate"] = img.getexif().get(0x0132)
                except PIL.UnidentifiedImageError:
                    invalid.append(i)

            elif row["extension"] in VIDEO_EXTENSIONS:
                try:
                    vid = cv2.VideoCapture(row['filepath'])
                    if vid.isOpened():
                        files.loc[i, "width"] = int(vid.get(cv2.CAP_PROP_FRAME_WIDTH))
                        files.loc[i, "height"] = int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT))
                    else:
                        invalid.append(i)
                    vid.release()
                except Exception:
                    invalid.append(i)

        # get filemodifydate as backup (videos, etc)
        files["filemodifydate"] = files["filepath"].apply(lambda x: datetime.fromtimestamp(Path(x).stat().st_mtime).strftime('%Y-%m-%d %H:%M:%S'))
        files["filemodifydate"] = pd.to_datetime(files["filemodifydate"]) + timedelta(hours=offset)
        try:
            # convert multiple string formats to datetime
            files['createdate'] = files['createdate'].replace(r'^\s*$', None, regex=True)
            files["createdate"] = files['createdate'].apply(lambda x: check_time(x) if isinstance(x, str) else x)
            files["createdate"] = pd.to_datetime(files["createdate"])
            # select createdate if exists, else choose filemodify date
            files["datetime"] = files['createdate'].fillna(files['filemodifydate'])
        except KeyError:
            files["datetime"] = files["filemodifydate"]

        # convert to datetime
        files["datetime"] = pd.to_datetime(files["datetime"])

    files = files.drop(index=invalid).reset_index(drop=True)

    if out_file:
        save_data(files, out_file)

    return files


class WorkingDirectory():
    """
    Set Working Directory and save file global variables.

    Constructor requires root working_directory
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, working_dir):
        if not isinstance(working_dir, PosixPath):
            working_dir = Path(working_dir)  # OS-agnostic path
        if not working_dir.is_dir():
            raise FileNotFoundError(f"The given directory: {working_dir}, does not exist.")

        self.basedir = working_dir / Path("Animl-Directory/")
        self.linkdir = self.basedir / Path("Sorted/")
        self.visdir = self.basedir / Path("Plots/")

        # Create directories if they do not already exist
        self.basedir.mkdir(exist_ok=True)

        # Assign specific file paths
        self.filemanifest = self.basedir / Path("FileManifest.csv")
        self.imageframes = self.basedir / Path("ImageFrames.csv")
        self.results = self.basedir / Path("Results.csv")
        self.predictions = self.basedir / Path("Predictions.csv")
        self.detections = self.basedir / Path("Detections.csv")
        self.mdraw = self.basedir / Path("MD_Raw.json")

    def activate_visdir(self):
        self.visdir.mkdir(exist_ok=True)

    def activate_linkdir(self):
        self.linkdir.mkdir(exist_ok=True)


def save_data(data: pd.DataFrame,
              out_file: str,
              prompt: bool = True) -> None:
    """
    Save data to given file.

    Args:
        data (pd.DataFrame): the dataframe to be saved
        out_file (str): full path to save file to
        prompt (bool): prompts the user to confirm overwrite

    Returns:
        None
    """
    if Path(out_file).is_file() and (prompt is True):
        prompt = "Output file exists, would you like to overwrite? y/n: "
        if input(prompt).lower() != "y":
            return
    else:
        if Path(out_file).parent.exists():
            data.to_csv(out_file, index=False)
        else:
            raise AssertionError('Cannot save, directory does not exis.')


def load_data(file: str) -> pd.DataFrame:
    """
    Load .csv File.

    Args:
        file (str): the full path of the file to load

    Returns:
        data extracted from the file. pd.dataframe form
    """
    if Path(file).suffix.lower() == ".csv":
        return pd.read_csv(file)
    else:
        raise AssertionError("Expecting a .csv file.")


def save_json(data: dict,
              out_file: str,
              prompt: bool = True) -> None:
    """
    Save data to a JSON file.

    Args:
        data (dict): the dictionary to be saved
        out_file (str): full path to save file to
        prompt (bool): prompt user to confirm overwrite

    Returns:
        None
    """
    if Path(out_file).is_file() and (prompt is True):
        prompt = "Output file exists, would you like to overwrite? y/n: "
        if input(prompt).lower() != "y":
            return
    with open(out_file, 'w') as f:
        json.dump(data, f, indent=4)


def load_json(file: str) -> dict:
    """
    Load data from a JSON file.

    Args:
        file (str): the full path of the file to load

    Returns:
        data extracted from the file. dict form
    """
    if Path(file).suffix.lower() == ".json":
        with open(file, 'r') as f:
            return json.load(f)
    else:
        raise AssertionError("Error. Expecting a .json file.")


def check_file(file: str, output_type: str = None) -> bool:
    """
    Check for files existence and prompt user if they want to load.

    Args:
        file (str): the full path of the file to check
        output_type (str): type of output file (e.g., "Manifest", "Detections")

    Returns:
        a boolean indicating whether a file was found and the user wants to load or not
    """

    if file is not None and Path(file).is_file():
        date = datetime.fromtimestamp(Path(file).stat().st_mtime).strftime('%Y-%m-%d %H:%M:%S')
        if output_type is None:
            output_type = "Output"
        prompt = f"{output_type} file already exists and was last modified {date}, would you like to load it? y/n: "
        response = input(prompt)
        if response.lower() == "y":
            return True
        elif response.lower() == "n":
            return False
        else:
            print("Invalid input, proceeding without loading file.")
    return False


def save_detection_checkpoint(checkpoint_path: str, results: dict) -> None:
    """
    Save a checkpoint of the detection results to a JSON file.

    Args:
        checkpoint_path (str): the path to the checkpoint file
        results (list): a list of detection results to save
    """
    # Back up any previous checkpoints, to protect against crashes while we're writing
    # the checkpoint file.
    checkpoint_tmp_path = None
    if Path(checkpoint_path).is_file():
        checkpoint_tmp_path = str(checkpoint_path) + '_tmp'
        copyfile(checkpoint_path, checkpoint_tmp_path)

    # Write the new checkpoint
    save_json({'images': results}, checkpoint_path, prompt=False)

    # Remove the backup checkpoint if it exists
    if checkpoint_tmp_path is not None:
        Path(checkpoint_tmp_path).unlink()


def active_times(manifest_dir,
                 depth: int = 1,
                 recursive: bool = True,
                 offset: int = 0) -> pd.DataFrame:
    """
    Get start and stop dates for each camera folder.

    Args:
        manifest_dir (str): either file manifest or directory of files to analyze
        depth (int): directory depth from which to split cameras
        recursive (bool): recursively search thhrough all child directories
        offset (int): add timezone offset in hours to datetime column

    Returns:
        times (pd.DataFrame): list of files with or without file modify dates
    """
    # from manifest file
    if isinstance(manifest_dir, str):
        if check_file(manifest_dir):
            files = load_data(manifest_dir)  # load_data(outfile) load file manifest

    # from manifest dataframe
    elif isinstance(manifest_dir, pd.DataFrame):
        # get time stamps if dne
        if "filemodifydate" not in manifest_dir.columns:
            files = manifest_dir
            files["filemodifydate"] = files["filepath"].apply(lambda x: datetime.fromtimestamp(Path(x).stat().st_mtime).strftime('%Y-%m-%d %H:%M:%S'))

    # from scratch
    elif Path(manifest_dir).is_dir():
        files = build_file_manifest(manifest_dir, exif=True, offset=offset, recursive=recursive)
    else:
        raise FileNotFoundError("Requires a file manifest or image directory.")

    files["camera"] = files["filepath"].apply(lambda x: Path(x).parts[depth])

    times = files.groupby("camera").agg({'filemodifydate': ['min', 'max']})

    return times


def sequence_calculation(manifest,
                         station_col: str,
                         sort_columns: list[str] = None,
                         file_col: str = "filepath",
                         maxdiff: int = 60):
    """
    Simple sequence calculation based on time differences between consecutive images from the same station.
    Unlike sequence_classification(), does not apply any classification or labeling to the sequences.

    Args:
        - manifest (pd.DataFrame): DataFrame containing image file information, including 'filepath' and 'datetime' columns
        - station_col (str): column name in the DataFrame representing the station or camera
        - sort_columns (list[str]): list of columns to sort by before calculating sequences. Defaults to None, which sorts by station_col and 'datetime'.
        - file_col (str): column name representing the file path. Defaults to "filepath".
        - maxdiff (int): maximum time difference in seconds between consecutive images to be considered part of the same sequence. Defaults to 60.
    """
    if not isinstance(station_col, str) or station_col == '':
        raise Exception("'station_col' must be a non-empty string")

    # Sanity check to verify that maxdiff is a positive number
    if not isinstance(maxdiff, (int, float)) or maxdiff < 0:
        raise Exception("'maxdiff' must be a number >= 0")

    if not {file_col}.issubset(manifest.columns):
        raise ValueError(f"DataFrame must contain '{file_col}' column.")

    if not {"datetime"}.issubset(manifest.columns):
        raise ValueError("DataFrame must contain 'datetime' column.")

    if sort_columns is None:
        sort_columns = [station_col, "datetime"]

    manifest['datetime'] = pd.to_datetime(manifest['datetime'], format="%Y-%m-%d %H:%M:%S")

    sort = manifest.sort_values(by=sort_columns).index
    manifest_sort = manifest.loc[sort].reset_index(drop=True)

    sequence_placeholder = np.zeros(len(manifest_sort))

    i = 0
    s = 0
    while i < len(manifest_sort):
        rows = [i]
        last_index = i+1

        while (last_index < len(manifest_sort) and not pd.isna(manifest_sort.loc[i, "datetime"]) and
               not pd.isna(manifest_sort.loc[last_index, "datetime"]) and
               manifest_sort.loc[last_index, station_col] == manifest_sort.loc[i, station_col] and
               (manifest_sort.loc[last_index, "datetime"] - manifest_sort.loc[i, "datetime"]).total_seconds() <= maxdiff):
            rows.append(last_index)
            last_index += 1

        sequence_placeholder[np.array(rows)] = int(s)

        i = last_index
        s += 1

    manifest_sort['sequence'] = sequence_placeholder

    return manifest_sort
